import click

from corgie.data_backends.base import get_data_backends, str_to_backend

# when adding a new backend type, include it here
from corgie.data_backends.cvbackend import CVDataBackend
from corgie.data_backends.base import DataBackendBase

# also update default backend if you want to
DEFAULT_DATA_BACKEND = 'cv'

pass_data_backend = click.make_pass_decorator(DataBackendBase)





import click

from corgie import scheduling
from corgie.log import logger as corgie_logger
from corgie.data_backends import pass_data_backend
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie import argparsers
from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option


class UpsampleJob(scheduling.Job):
    def __init__(self, src_layer, dst_layer, mip_start, mip_end,
                 bcube, chunk_xy, chunk_z, mips_per_task):
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mip_start = mip_start
        self.mip_end = mip_end
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.mips_per_task = mips_per_task

        super().__init__()

    def task_generator(self):
        for mip in range(self.mip_start, self.mip_end, -self.mips_per_task):
            this_mip_start = mip
            this_mip_end = min(self.mip_end, mip + self.mips_per_task)
            chunks = self.dst_layer.break_bcube_into_chunks(
                    bcube=self.bcube,
                    chunk_xy=self.chunk_xy,
                    chunk_z=self.chunk_z,
                    mip=this_mip_start)

            tasks = [UpsampleTask(self.src_layer,
                                    self.dst_layer,
                                    this_mip_start,
                                    this_mip_end,
                                    input_chunk) for input_chunk in chunks]
            print ("Yielding upsample tasks for bcube: {}, MIPs: {}-{}".format(
                self.bcube, this_mip_start, this_mip_end))

            yield tasks

            if mip == self.mip_start:
                self.src_layer = self.dst_layer

            # if not the last iteration
            if mip + self.mips_per_task < self.mip_end:
                yield scheduling.wait_until_done


class UpsampleTask(scheduling.Task):
    def __init__(self, src_layer, dst_layer, mip_start, mip_end,
                 bcube):
        super().__init__()
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mip_start = mip_start
        self.mip_end = mip_end
        self.bcube = bcube

    def execute(self):
        src_data = self.src_layer.read(bcube=self.bcube, mip=self.mip_start)
        # How to upsample depends on layer type.
        # Images are avg pooled, masks are max pooled, segmentation is...
        upsampler = self.src_layer.get_upsampler()
        for mip in range(self.mip_start, self.mip_end, -1):
            dst_data = upsampler(src_data)
            self.dst_layer.write(dst_data, bcube=self.bcube, mip=mip-1)
            src_data = dst_data


@click.command()
@corgie_optgroup('Layer Parameterd')
@corgie_option('--src_layer_spec',  '-s', nargs=1,
        type=str, required=True,
        help='Specification for the source layer. ' + \
                LAYER_HELP_STR)

@corgie_option('--dst_layer_spec',  '-s', nargs=1,
        type=str, required=False,
        help= "Specification for the destination layer. "+ \
                "Refer to 'src_layer_spec' for parameter format." + \
                " DEFAULT: Same as src_layer_spec")

@corgie_optgroup('Upsample parameters')
@corgie_option('--mip_start',  '-m', nargs=1, type=int, required=True)
@corgie_option('--mip_end',    '-e', nargs=1, type=int, required=True)
@corgie_option('--chunk_xy',   '-c', nargs=1, type=int, default=2048)
@corgie_option('--chunk_z',          nargs=1, type=int, default=1)
@corgie_option('--mips_per_task',    nargs=1, type=int, default=3)

@corgie_optgroup('Data Region Specification')
@corgie_option('--start_coord',      nargs=1, type=str, required=True)
@corgie_option('--end_coord',        nargs=1, type=str, required=True)
@corgie_option('--coord_mip',        nargs=1, type=int, default=0)
@click.pass_context
def upsample(ctx, src_layer_spec, dst_layer_spec, mip_start,
        mip_end, chunk_xy, chunk_z, mips_per_task, start_coord,
        end_coord, coord_mip):
    scheduler = ctx.obj['scheduler']
    corgie_logger.debug("Setting up Source and Destination layers...")

    src_layer = create_layer_from_spec(src_layer_spec,
            caller_name='src layer',
            readonly=True)

    if dst_layer_spec is None:
        corgie_logger.info("Destination layer not specified. Using Source layer "
                "as Destination.")
        dst_layer = src_layer
        dst_layer.readonly = False
    else:
        dst_layer = create_layer_from_spec(dst_layer_spec,
            caller_name='dst_layer layer',
            readonly=False,
            reference=src_layer, chunk_z=chunk_z, overwrite=True)
    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)
    upsample_job = UpsampleJob(src_layer, dst_layer,
                                   mip_start, mip_end,
                                   bcube=bcube,
                                   chunk_xy=chunk_xy,
                                   chunk_z=chunk_z,
                                   mips_per_task=mips_per_task)

    # create scheduler and execute the job
    scheduler.register_job(upsample_job, job_name="upsample")
    scheduler.execute_until_completion()
    result_report = f"Upsampled {src_layer} from {mip_start} to {mip_end}. Result in {dst_layer}"
    corgie_logger.info(result_report)



import copy

import torch

from corgie import constants, exceptions

from corgie.log import logger as corgie_logger
from corgie.boundingcube import BoundingCube
from corgie.layers.base import register_layer_type, BaseLayerType
from corgie import helpers

class VolumetricLayer(BaseLayerType):
    def __init__(self, data_mip=None, **kwargs):
        super().__init__(**kwargs)
        self.data_mip = data_mip

    def read(self, bcube, mip, **kwargs):
        indexed_bcube = self.indexing_scheme(bcube, mip, kwargs)
        if self.data_mip is not None:
            if mip <= self.data_mip:
                result_data_mip = super().read(bcube=indexed_bcube,
                        mip=self.data_mip, **kwargs)
                result = result_data_mip
                for _ in range(mip, self.data_mip):
                    result = self.get_upsampler()(result)
            else:
                raise exceptions.CorgieException(f"Attempting to read from higher MIP (MIP {mip}) "
                        f"than the declared data MIP (MIP {self.data_mip}) for {str(self)}")
        else:
            result = super().read(bcube=indexed_bcube, mip=mip, **kwargs)

        return result

    def write(self, data_tens, bcube, mip, **kwargs):
        indexed_bcube = self.indexing_scheme(bcube, mip, kwargs)
        super().write(data_tens=data_tens, bcube=indexed_bcube, mip=mip, **kwargs)

    def indexing_scheme(self, bcube, mip, kwargs):
        return bcube

    def break_bcube_into_chunks(self, bcube, chunk_xy, chunk_z,
            mip, flatten=True, chunk_xy_step=None, chunk_z_step=None,
            **kwargs):
        """Default breaking up of a bcube into smaller bcubes (chunks).
        Returns a list of chunks
        Args:
           bcube: BoundingBox for region to be broken into chunks
           chunk_size: tuple for dimensions of chunk that bbox will be broken into
           mip: int for MIP level at which chunk_xy is dspecified
        """
        indexed_bcube = self.indexing_scheme(bcube, mip, kwargs)

        x_range = indexed_bcube.x_range(mip=mip)
        y_range = indexed_bcube.y_range(mip=mip)
        z_range = indexed_bcube.z_range()

        if chunk_xy_step is None:
            chunk_xy_step = chunk_xy
        if chunk_z_step is None:
            chunk_z_step = chunk_z

        xy_chunks = []
        flat_chunks = []
        for zs in range(z_range[0], z_range[1], chunk_z_step):
            xy_chunks.append([])
            for xs in range(x_range[0], x_range[1], chunk_xy_step):
                xy_chunks[-1].append([])
                for ys in range(y_range[0], y_range[1], chunk_xy_step):
                    chunk = BoundingCube(xs, xs + chunk_xy,
                                         ys, ys + chunk_xy,
                                         zs, zs + chunk_z,
                                         mip=mip)

                    xy_chunks[-1][-1].append(chunk)
                    flat_chunks.append(chunk)

        if flatten:
            return flat_chunks
        else:
            return xy_chunks


@register_layer_type("img")
class ImgLayer(VolumetricLayer):
    def __init__(self, *args, num_channels=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_channels = num_channels

    def get_downsampler(self):
        def downsampler(data_tens):
            return torch.nn.functional.interpolate(data_tens.float(),
                    mode='bilinear',
                    scale_factor=1/2,
                    align_corners=False,
                    recompute_scale_factor=False)
        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            return torch.nn.functional.interpolate(data_tens.float(),
                    mode='bilinear',
                    scale_factor=2.0,
                    align_corners=False,
                    recompute_scale_factor=False)
        return upsampler

    def get_num_channels(self, *args, **kwargs):
        return self.num_channels

    def get_default_data_type(self):
        return 'uint8'

@register_layer_type("field")
class FieldLayer(VolumetricLayer):
    def __init__(self, *args, num_channels=2, **kwargs):
        if num_channels != 2:
            raise exceptions.ArgumentError("Field layer 'num_channels'",
                    "Field layer must have 2 channels. 'num_channels' provided: {}".format(
                        num_channels
                        ))
        super().__init__(*args, **kwargs)

    def get_downsampler(self):
        def downsampler(data_tens):

            downs_data = torch.nn.functional.interpolate(data_tens.float(),
                                mode='bilinear',
                                scale_factor=1/2,
                                align_corners=False,
                    recompute_scale_factor=False)
            return downs_data * 0.5

        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            ups_data = torch.nn.functional.interpolate(data_tens.float(),
                                mode='bilinear',
                                scale_factor=2.0,
                                align_corners=False,
                    recompute_scale_factor=False)
            return ups_data * 2

        return upsampler

    def get_num_channels(self, *args, **kwargs):
        return 2

    def get_default_data_type(self):
        return 'float32'


@register_layer_type("mask")
class MaskLayer(VolumetricLayer):
    def __init__(self, binarization=None,
            num_channels=1, **kwargs):
        self.binarizer = helpers.Binarizer(binarization)
        if num_channels != 1:
            raise exceptions.ArgumentError("Mask layer 'num_channels'",
                    "Mask layer must have 1 channels. 'num_channels' provided: {}".format(
                        num_channels
                        ))
        super().__init__(**kwargs)

    def read(self, **kwargs):
        data_tens = super().read(**kwargs)
        data_bin = self.binarizer(data_tens)
        return data_bin

    def get_downsampler(self):
        def downsampler(data_tens):
            return torch.nn.functional.interpolate(data_tens.float(),
                    mode='nearest',
                    scale_factor=1/2,
                    recompute_scale_factor=False)
        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            return torch.nn.functional.interpolate(data_tens.float(),
                    mode='nearest',
                    scale_factor=2.0,
                    recompute_scale_factor=False)
        return upsampler

    def get_num_channels(self, *args, **kwargs):
        return 1

    def get_default_data_type(self):
        return 'uint8'

@register_layer_type("section_value")
class SectionValueLayer(VolumetricLayer):
    def __init__(self, *args, num_channels=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_channels = num_channels

    # TODO: insert custom indexing here.

    def get_num_channels(self, *args, **kwargs):
        return 1

    def indexing_scheme(self, bcube, mip, kwargs):
        new_bcube = copy.deepcopy(bcube)
        if 'channel_start' in kwargs and 'channel_end' in kwargs:
            channel_start = kwargs['channel_start']
            channel_end = kwargs['channel_end']
            del kwargs['channel_start'], kwargs['channel_end']
        else:
            channel_start = 0
            channel_end = self.num_channels

        new_bcube.reset_coords(channel_start, channel_end, 0, 1, mip=mip)
        return new_bcube

    def supports_voxel_offset(self):
        return False

    def supports_chunking(self):
        return False

    def get_default_data_type(self):
        return 'float32'



import json
import copy
from math import floor, ceil
import numpy as np

from corgie.helpers import crop

def get_bcube_from_coords(start_coord, end_coord, coord_mip,
        cant_be_empty=True):
    xs, ys, zs = [int(i) for i in start_coord.split(',')]
    xe, ye, ze = [int(i) for i in end_coord.split(',')]
    bcube = BoundingCube(xs, xe, ys, ye, zs, ze, coord_mip)

    if cant_be_empty and bcube.area() * bcube.z_size() == 0:
        raise Exception("Attempted creation of an empty bounding "
                "when 'cant_be_empty' flag is set to True")

    return bcube

class BoundingCube:
    def __init__(self, xs, xe, ys, ye, zs, ze, mip):
        self.m0_x = (None, None)
        self.m0_y = (None, None)
        self.z = (None, None)
        self.reset_coords(xs, xe, ys, ye, zs, ze, mip=mip)

    # TODO
    # def contains(self, other):
    # def insets(self, other, mip):

    def get_bounding_pts(self):
        return (self.m0_x[0], self.m0_y[0], self.z[0]), \
               (self.m0_x[1], self.m0_y[1], self.z[1])

    def contains(self, other):
        if self.m0_y[1] < other.m0_y[1]:
            return False
        if self.m0_x[1] < other.m0_x[1]:
            return False
        if self.z[1] < other.z[1]:
            return False

        if other.m0_x[0] < self.m0_x[0]:
            return False
        if other.m0_y[0] < self.m0_y[0]:
            return False
        if other.z[0] < self.z[0]:
            return False

        return True


    # TODO: delete?
    def intersects(self, other):
        assert type(other) == type(self)
        if other.m0_x[1] < self.m0_x[0]:
            return False
        if other.m0_y[1] < self.m0_y[0]:
            return False
        if self.m0_x[1] < other.m0_x[0]:
            return False
        if self.m0_y[1] < other.m0_y[0]:
            return False
        if self.z[1] < other.z[0]:
            return False
        if other.z[1] < self.z[0]:
            return False
        return True

    def reset_coords(self, xs=None, xe=None,
            ys=None, ye=None, zs=None, ze=None, mip=0, in_place=True):

        if in_place:
            target_cube = self
        else:
            target_cube = copy.deepcopy(self)
        scale_factor = 2**mip
        if xs is not None:
            target_cube.m0_x =  (int(xs * scale_factor), target_cube.m0_x[1])
        if xe is not None:
            target_cube.m0_x = (target_cube.m0_x[0], int(xe * scale_factor))

        if ys is not None:
            target_cube.m0_y =  (int(ys * scale_factor), target_cube.m0_y[1])
        if ye is not None:
            target_cube.m0_y = (target_cube.m0_y[0], int(ye * scale_factor))

        if zs is not None:
            target_cube.z =  (int(zs), target_cube.z[1])
        if ze is not None:
            target_cube.z = (target_cube.z[0], int(ze))

        return target_cube

    def get_offset(self, mip=0):
        scale_factor = 2**mip
        return (self.m0_x[0] / scale_factor + self.x_size(mip=0) / 2 / scale_factor,
                self.m0_y[0] / scale_factor + self.y_size(mip=0) / 2 / scale_factor)

    def x_range(self, mip):
        scale_factor = 2**mip
        xs = floor(self.m0_x[0] / scale_factor)
        xe = ceil(self.m0_x[1] / scale_factor)
        return [xs, xe]

    def y_range(self, mip):
        scale_factor = 2**mip
        ys = floor(self.m0_y[0] / scale_factor)
        ye = ceil(self.m0_y[1] / scale_factor)
        return [ys, ye]

    def z_range(self):
        return list(copy.deepcopy(self.z))

    def area(self, mip=0):
        x_size = self.x_size(mip)
        y_size = self.y_size(mip)
        return x_size * y_size

    def x_size(self, mip):
        x_range = self.x_range(mip)
        return int(x_range[1] - x_range[0])

    def y_size(self, mip):
        y_range = self.y_range(mip)
        return int(y_range[1] - y_range[0])

    def z_size(self):
        return int(self.z[1] - self.z[0])

    @property
    def size(self, mip=0):
        return self.x_size(mip=mip), self.y_size(mip=mip), self.z_size()

    def crop(self, crop_xy, mip):
        scale_factor = 2**mip
        m0_crop_xy = crop_xy * scale_factor
        self.set_m0(self.m0_x[0] + m0_crop_xy,
                    self.m0_x[1] - m0_crop_xy,
                    self.m0_y[0] + m0_crop_xy,
                    self.m0_y[1] - m0_crop_xy)

    def uncrop(self, crop_xy, mip):
        """Uncrop the bounding box by crop_xy at given MIP level
        """
        scale_factor = 2**mip
        m0_crop_xy = crop_xy * scale_factor
        result = self.clone()
        result.reset_coords(xs=self.m0_x[0] - m0_crop_xy,
                        xe=self.m0_x[1] + m0_crop_xy,
                        ys=self.m0_y[0] - m0_crop_xy,
                        ye=self.m0_y[1] + m0_crop_xy,
                        mip=0)
        return result

    def zeros(self, mip):
        return np.zeros((self.x_size(mip), self.y_size(mip), self.z_size()),
                dtype=np.float32)

    def x_res_displacement(self, d_pixels, mip):
        disp_prop = d_pixels / self.x_size(mip=0)
        result = np.full((self.x_size(mip), self.y_size(mip)), disp_prop, dtype=np.float32)
        return result

    def y_res_displacement(self, d_pixels, mip):
        disp_prop = d_pixels / self.y_size(mip=0)
        result = np.full((self.x_size(mip), self.y_size(mip)), disp_prop, dtype=np.float32)
        return result

    def spoof_x_y_residual(self, x_d, y_d, mip, crop_amount=0):
        x_res = crop(self.x_res_displacement(x_d, mip=mip), crop_amount)
        y_res = crop(self.y_res_displacement(y_d, mip=mip), crop_amount)
        result = np.stack((x_res, y_res), axis=2)
        result = np.expand_dims(result, 0)
        return result

    def __eq__(self, x):
        if isinstance(x, BoundingBox):
            return (self.m0_x == x.m0_x) and (self.m0_y == x.m0_y) and \
                    (self.z == x.z)
        return False

    def __str__(self, mip=0):
        return "[MIP {}] {}, {}, {}".format(
                mip, self.x_range(mip), self.y_range(mip),
                self.z_range())

    def __repr__(self):
        return self.__str__(mip=0)

    def translate_v1(self, x=0, y=0, z=0, mip=0):
        assert isinstance(x, int)
        assert isinstance(y, int)
        assert isinstance(z, int)

        scale_factor = 2**mip

        return BoundingCube(xs=self.m0_x[0] + x,
                           xe=self.m0_x[1] + x,
                           ys=self.m0_y[0] + y,
                           ye=self.m0_y[1] + y,
                           zs=self.z[0] + z,
                           ze=self.z[1] + z,
                           mip=mip)

    def clone(self):
        return copy.deepcopy(self)

    def translate(self, z_offset=0, x_offset=0, y_offset=0,
            mip=0):
        x_range = self.x_range(mip=mip)
        y_range = self.y_range(mip=mip)
        z_range = self.z_range()
        return BoundingCube(
                xs=x_range[0] + x_offset,
                xe=x_range[1] + x_offset,
                ys=y_range[0] + y_offset,
                ye=y_range[1] + y_offset,
                zs=z_range[0] + z_offset,
                ze=z_range[1] + z_offset,
                mip=mip
            )


    def to_slices(self, zs, ze=None, mip=0):
        x_range = self.x_range(mip=mip)
        y_range = self.y_range(mip=mip)
        return slice(*x_range), slice(*y_range), slice(*self.z)





import json
import click

from click_option_group import optgroup

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, str_to_layer_type, \
        DEFAULT_LAYER_TYPE
from corgie.data_backends import get_data_backends, str_to_backend, \
        DEFAULT_DATA_BACKEND
from corgie import exceptions

from corgie.stack import Stack

corgie_optgroup = optgroup.group

def corgie_option(*args, **kwargs):
    return optgroup.option(*args, show_default=True, **kwargs)

LAYER_HELP_STR = """
   [required]\n
   Format: JSON *string*. \n
   Optional keys: \n
   \t "type": str from {} ; \n
   \t "name": str, DEFAULT -- same as layer type. """\
   """**Required** if given more than 1 layer of the same type ; \n
   \t "data_backend": str from {} ; \n
   \t "layer_args": a dictionary with additional layer parameters,
   eg binarization scheme for masks data_mip_ranges, etc; \n
   Required keys: "path". \n
   """.format(get_layer_types(), get_data_backends())

   #\t "data_mip_ranges": array of [int, int] tupples specifying MIP ranges where the layer has data. DEFAULT -- all MIPs have data.\n
def create_layer_from_spec(spec_str, reference=None, caller_name=None,
        allowed_types=None, default_type=None, **kwargs):
    if default_type is None:
        default_type = DEFAULT_LAYER_TYPE
    default_param_dict = {
            "path": None,
            "name": None,
            "type": default_type,
            "data_backend": DEFAULT_DATA_BACKEND,
            "args": {}
            }
    param_dict = json.loads(spec_str)
    for k in param_dict.keys():
        if k not in default_param_dict:
            raise exceptions.CorgieException(f"Unkown layer parameter '{k}'")
    params = {**default_param_dict, **param_dict}

    if params["path"] is None:
        arg_spec = '"path" key in layer specification'
        if caller_name is not None:
            arg_spec += ' of {}'.format(caller_name)
        raise exceptions.ArgumentError(arg_spec, 'not given')

    layer_path = params["path"]
    layer_type = params["type"]
    layer_args = params["args"]
    data_backend = params["data_backend"]
    corgie_logger.debug("Parsing layer path: {}".format(
        layer_path
        ))

    if allowed_types is not None and layer_type not in allowed_types:
        raise exceptions.ArgumentError(arg_spec, 'must be of type in {allowed_types}')

    backend = str_to_backend(data_backend)()

    layer = backend.create_layer(path=layer_path, layer_type=layer_type,
            reference=reference, layer_args=layer_args,
            **kwargs)

    name = params["name"]
    if name is None:
        name = layer.get_layer_type()
    layer.name = name

    return layer


def create_stack_from_spec(spec_str_list, name, reference=None, readonly=False,
        **kwargs):
    stack = None
    if len(spec_str_list) == 0:
        if reference is not None:
            stack = Stack(name=name)
            for l in reference.get_layers():
                l.readonly = readonly
                stack.add_layer(l)
    else:
        stack = Stack(name=name)
        layer_list = [create_layer_from_spec(s, readonly=readonly, **kwargs) \
                for s in spec_str_list]
        for l in layer_list:
            stack.add_layer(l)
    return stack




import torch
import numpy as np

from corgie import constants, exceptions
from corgie import helpers


STR_TO_LTYPE_DICT  = dict()


def register_layer_type(layer_type_name):
    def register_layer_fn(layer_type):
        STR_TO_LTYPE_DICT[layer_type_name] = layer_type
        def get_layer_type_name(obj):
            return layer_type_name
        layer_type.get_layer_type = get_layer_type_name
        return layer_type

    return register_layer_fn


def str_to_layer_type(s):
    global STR_TO_LTYPE_DICT
    return STR_TO_LTYPE_DICT[s]


def get_layer_types():
    return list(STR_TO_LTYPE_DICT.keys())


class BaseLayerType:
    def __init__(self, name=None, device='cpu', readonly=False, **kwargs):
        super().__init__(**kwargs)
        self.device = device
        self.readonly = readonly
        self.name = name

    def read(self, dtype=None, **kwargs):
        data_np = self.read_backend(**kwargs)
        # TODO: if np type is unit32, convert it to int64
        if data_np.dtype == np.uint32:
            data_np = data_np.astype(np.int64)
        data_tens = torch.as_tensor(data_np, device=self.device)
        data_tens = helpers.cast_tensor_type(data_tens, dtype)
        return data_tens

    def write(self, data_tens, **kwargs):
        if self.readonly:
            raise Exception("Attempting to write into a readonly layer {}".format(str(self)))
        data_tens = helpers.expand_to_dims(data_tens, 4)
        if data_tens.dtype == torch.float64:
            data_tens = data_tens.float()
        #if data_tens.dtype in [torch.float32, torch.float64]:
            #import pdb; pdb.set_trace()
            #if self.get_data_type() in ['uint8']:
            #    data_tens = data_tens * 255
            #if self.get_data_type() not in ['float32', 'float64', 'float']:
            #    raise Exception("Unknown conversiotn between float and int")

        data_np = data_tens.data.cpu().numpy().astype(
                self.get_data_type()
                )

        self.write_backend(data_np, **kwargs)

    def supports_voxel_offset(self):
        return True

    def supports_chunking(self):
        return True

    def get_downsampler(self, *kargs, **kwargs):
        raise NotImplementedError

    def get_upsampler(self, *kargs, **kwargs):
        raise NotImplementedError

    def get_num_channels(self, *kargs, **kwargs):
        raise NotImplementedError



from dataclasses import dataclass


class Binarizer:
    def __init__(self, binarization):
        self.bin = binarization
    def __call__(self, tens):
        if self.bin is None:
            return tens
        elif self.bin[0] == 'neq':
            return tens != self.bin[1]
        elif self.bin[0] == 'eq':
            return tens == self.bin[1]

class PartialSpecification:
    def __init__(self, f, **kwargs):
        self.f = f
        self.constr_kwargs = kwargs

    def __call__(self, **kwargs):
        return self.f(**self.constr_kwargs, **kwargs)


@dataclass
class Translation:
    x: float
    y: float

def crop(**kwargs):
    raise NotImplementedError

def expand_to_dims(tens, dims):
    tens_dims = len(tens.shape)
    assert (tens_dims) <= dims
    tens = tens[(None, ) * (dims - tens_dims)]
    return tens

def cast_tensor_type(tens, dtype):
    '''
        tens: pytorch tens
        dtype: string, eg 'float', 'int', 'byte'
    '''
    if dtype is not None:
        assert hasattr(tens, dtype)
        return getattr(tens, dtype)()
    else:
        return tens

def read_mask_list(mask_list, bcube, mip):
    result = None
    for m in mask_list:
        this_data = m.read(bcube=bcube, mip=mip)
        if result is None:
            result = this_data
        else:
            result += this_data
            result = result > 0

    return result

def crop(data, c):
    if c == 0:
        return data
    else:
        if data.shape[-1] == data.shape[-2]:
            return data[..., c:-c, c:-c]
        elif data.shape[-2] == data.shape[-3] and data.shape[-1] == 2: #field
            return data[..., c:-c, c:-c, :]




import torch

from corgie import exceptions

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, str_to_layer_type

#TODO: DEFNITELY need a full blown cache.

STR_TO_BACKEND_DICT = {}


def str_to_backend(s):
    global STR_TO_BACKEND_DICT
    return STR_TO_BACKEND_DICT[s]


def get_data_backends():
    return list(STR_TO_BACKEND_DICT.keys())


def register_backend(name):
    def register_backend_fn(cls):
        global STR_TO_BACKEND_DICT
        STR_TO_BACKEND_DICT[name] = cls
        return cls
    return register_backend_fn


class DataBackendBase:
    layer_constr_dict = {n: None for n in get_layer_types()}
    default_device = None

    def __init__(self, *kargs, device=None, **kwargs):
        if device is not None:
            self.device = device
        else:
            self.device = self.default_device
        super().__init__(*kargs, **kwargs)

    def create_layer(self, path, layer_type=None, reference=None, layer_args={}, **kwargs):
        if layer_type not in self.layer_constr_dict:
            raise exceptions.CorgieException("Layer type {} is not \
                    defined".format(layer_type))
        if self.layer_constr_dict[layer_type] is None:
            raise exceptions.CorgieException("Layer type {} is not \
                    implemented for {} backend".format(layer_type, type(self)))

        corgie_logger.debug("Creating layer '{}' on device '{}' with reference '{}'...".format(
                path, self.device, reference
                ))
        layer = self.layer_constr_dict[layer_type](path=path, device=self.device,
                reference=reference,
                backend=self,
                **layer_args,
                **kwargs
                )
        corgie_logger.debug("Done")
        return layer

    @classmethod
    def register_layer_type_backend(cls, layer_type_name):
        # This is a decorator for including
        assert layer_type_name in cls.layer_constr_dict

        def register_fn(layer):
            cls.layer_constr_dict[layer_type_name] = layer
            return layer

        return register_fn


class BaseLayerBackend:
    def __init__(self, dtype=None, **kwargs):
        super().__init__(**kwargs)
        self.dtype = dtype

    def get_data_type(self):
        return self.dtype

    def get_sublayer(obj, *kargs, **kwargs):
        raise Exception("layer type backend must implement "
                "'get_sublayer' function")

        #$def get_data_type(self):
    #    raise Exception("layer type backend must implement "
    #            "'get_data_type' function")

    def read_backend(self, *kargs, **kwargs):
        raise Exception("layer type backend must implement "
                "'read_backend' function")

    def write_backend(self, *kargs, **kwargs):
        raise Exception("layer type backend must implement"
                "'write_backend' function")




import click
from copy import deepcopy

from corgie import scheduling, argparsers, helpers, stack

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option, \
        create_stack_from_spec

from corgie.cli.render import RenderJob
from corgie.cli.copy import CopyJob
from corgie.cli.compute_field import ComputeFieldJob

class AlignBlockJob(scheduling.Job):
    def __init__(self, src_stack, tgt_stack, dst_stack, cf_method, render_method,
                 bcube, copy_start=True, backward=False, suffix=None):
        self.src_stack = src_stack
        self.tgt_stack = tgt_stack
        self.dst_stack = dst_stack
        self.bcube = bcube

        self.cf_method = cf_method
        self.render_method = render_method

        self.copy_start = copy_start
        self.backward = backward
        self.suffix = suffix

        super().__init__()

    def task_generator(self):

        if not self.backward:
            z_start = self.bcube.z_range()[0]
            z_end = self.bcube.z_range()[1]
            z_step = 1
        else:
            z_start = self.bcube.z_range()[1]
            z_end = self.bcube.z_range()[0]
            z_step = -1
        seethrough_offset = -z_step

        start_sec_bcube = self.bcube.reset_coords(
                zs=z_start, ze=z_start + 1, in_place=False)
        if self.copy_start:
            render_job = self.render_method(
                    src_stack=self.src_stack,
                    dst_stack=self.dst_stack,
                    bcube=start_sec_bcube,
                    seethrough=False,
                    blackout_masks=False)

            yield from render_job.task_generator
            yield scheduling.wait_until_done
        src_bcube = start_sec_bcube

        align_field_layer = self.dst_stack.create_sublayer(f'align_field{self.suffix}',
                layer_type='field', overwrite=True)

        for z in range(z_start + z_step, z_end + z_step, z_step):
            tgt_bcube = src_bcube
            src_bcube = self.bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            compute_field_job = self.cf_method(
                    src_stack=self.src_stack,
                    tgt_stack=self.dst_stack,
                    bcube=src_bcube,
                    tgt_z_offset=-z_step,
                    dst_layer=align_field_layer)

            yield from compute_field_job.task_generator
            yield scheduling.wait_until_done

            render_job = self.render_method(
                    src_stack=self.src_stack,
                    dst_stack=self.dst_stack,
                    bcube=src_bcube,
                    additional_fields=[align_field_layer],
                    seethrough=True,
                    blackout_masks=True,
                    seethrough_offset=seethrough_offset
                    )

            yield from render_job.task_generator
            yield scheduling.wait_until_done


@click.command()
# Layers
@corgie_optgroup('Layer Parameters')

@corgie_option('--src_layer_spec',  '-s', nargs=1,
        type=str, required=True, multiple=True,
        help='Source layer spec. Use multiple times to include all masks, fields, images. ' + \
                LAYER_HELP_STR)
#
@corgie_option('--tgt_layer_spec', '-t', nargs=1,
        type=str, required=False, multiple=True,
        help='Target layer spec. Use multiple times to include all masks, fields, images. \n'\
                'DEFAULT: Same as source layers')

@corgie_option('--dst_folder',  nargs=1, type=str, required=True,
        help="Folder where aligned stack will go")

@corgie_option('--suffix',              nargs=1, type=str,  default=None)

@corgie_optgroup('Render Method Specification')
#@corgie_option('--seethrough_masks',    nargs=1, type=bool, default=False)
#@corgie_option('--seethrough_misalign', nargs=1, type=bool, default=False)
@corgie_option('--render_pad',          nargs=1, type=int,  default=512)
@corgie_option('--render_chunk_xy',     nargs=1, type=int,  default=1024)

@corgie_optgroup('Compute Field Method Specification')
@corgie_option('--processor_spec',      nargs=1, type=str, required=True,
        multiple=True)
@corgie_option('--chunk_xy',      '-c', nargs=1, type=int, default=1024)
@corgie_option('--pad',                 nargs=1, type=int, default=256)
@corgie_option('--crop',                nargs=1, type=int, default=None)
@corgie_option('--processor_mip', '-m', nargs=1, type=int, required=True,
        multiple=True)
@corgie_option('--copy_start/--no_copy_start',             default=True)
@corgie_option('--mode', type=click.Choice(['forward', 'backword', 'bidirectional']),
        default='forward')

@corgie_optgroup('Data Region Specification')
@corgie_option('--start_coord',        nargs=1, type=str, required=True)
@corgie_option('--end_coord',          nargs=1, type=str, required=True)
@corgie_option('--coord_mip',          nargs=1, type=int, default=0)

@click.pass_context
def align_block(ctx, src_layer_spec, tgt_layer_spec, dst_folder, render_pad, render_chunk_xy,
        processor_spec, pad, crop, processor_mip, chunk_xy, start_coord, end_coord, coord_mip,
        suffix, copy_start, mode, chunk_z=1):
    scheduler = ctx.obj['scheduler']

    if suffix is None:
        suffix = '_aligned'
    else:
        suffix = f"_{suffix}"

    if crop is None:
        crop = pad
    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec,
            name='src', readonly=True)

    tgt_stack = create_stack_from_spec(tgt_layer_spec,
            name='tgt', readonly=True, reference=src_stack)

    dst_stack = stack.create_stack_from_reference(reference_stack=src_stack,
            folder=dst_folder, name="dst", types=["img", "mask"], readonly=False,
            suffix=suffix, overwrite=True)

    render_method = helpers.PartialSpecification(
            f=RenderJob,
            pad=render_pad,
            chunk_xy=render_chunk_xy,
            chunk_z=1,
            render_masks=False,
            mip=min(processor_mip)
            )

    cf_method = helpers.PartialSpecification(
            f=ComputeFieldJob,
            pad=pad,
            crop=crop,
            processor_mip=processor_mip,
            processor_spec=processor_spec,
            chunk_xy=chunk_xy,
            chunk_z=1
            )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if mode == 'bidirectional':
        z_mid = (bcube.z_range()[1] + bcube.z_range()[0]) // 2
        bcube_back = bcube.reset_coords(ze=z_mid, in_place=False)
        bcube_forv = bcube.reset_coords(zs=z_mid, in_place=False)

        align_block_job_back = AlignBlockJob(src_stack=src_stack,
                                    tgt_stack=tgt_stack,
                                    dst_stack=dst_stack,
                                    bcube=bcube_back,
                                    render_method=render_method,
                                    cf_method=cf_method,
                                    suffix=suffix,
                                    copy_start=copy_start,
                                    backward=True)
        scheduler.register_job(align_block_job_back, job_name="Backward Align Block {}".format(bcube))

        align_block_job_forv = AlignBlockJob(src_stack=src_stack,
                                    tgt_stack=tgt_stack,
                                    dst_stack=deepcopy(dst_stack),
                                    bcube=bcube_forv,
                                    render_method=render_method,
                                    cf_method=cf_method,
                                    suffix=suffix,
                                    copy_start=True,
                                    backward=False)
        scheduler.register_job(align_block_job_forv, job_name="Forward Align Block {}".format(bcube))
    else:
        align_block_job = AlignBlockJob(src_stack=src_stack,
                                        tgt_stack=tgt_stack,
                                        dst_stack=dst_stack,
                                        bcube=bcube,
                                        render_method=render_method,
                                        cf_method=cf_method,
                                        suffix=suffix,
                                        copy_start=copy_start,
                                        backward=mode=='backward')

        # create scheduler and execute the job
        scheduler.register_job(align_block_job, job_name="Align Block {}".format(bcube))

    scheduler.execute_until_completion()
    result_report = f"Aligned layers {[str(l) for l in src_stack.get_layers_of_type('img')]}. " \
            f"Results in {[str(l) for l in dst_stack.get_layers_of_type('img')]}"
    corgie_logger.info(result_report)






from corgie.layers.base import get_layer_types, str_to_layer_type

# when adding a new layer type, include it here
from corgie.layers.volumetric_layers import FieldLayer, ImgLayer, MaskLayer, \
        SectionValueLayer

# also update the default layer type if you want to
DEFAULT_LAYER_TYPE = 'img'



import click
import procspec

from corgie import scheduling, argparsers, helpers

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option, \
        create_stack_from_spec

from corgie.cli.common import ChunkedJob


class ApplyProcessorJob(scheduling.Job):
    def __init__(self, src_stack, dst_layer,
            chunk_xy, chunk_z, processor_spec, processor_mip,
            pad, crop,
            bcube, blend_xy=0):

        self.src_stack = src_stack
        self.dst_layer = dst_layer
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.blend_xy = blend_xy
        self.pad = pad
        self.crop = crop
        self.bcube = bcube
        self.processor_spec = processor_spec
        self.processor_mip = processor_mip
        if isinstance(self.processor_spec, str):
            self.processor_spec = [self.processor_spec]
        if isinstance(self.processor_mip, int):
            self.processor_mip = [self.processor_mip]

        if len(self.processor_mip) != len(self.processor_spec):
            raise exceptions.CorgieException("The number of processors doesn't "
                    "match the number of specified processor MIPs")

        super().__init__()

    def task_generator(self):
        for i in range(len(self.processor_spec)):
            this_proc = self.processor_spec[i]
            this_proc_mip = self.processor_mip[i]
            is_last_proc = i == len(self.processor_spec) - 1

            this_task = helpers.PartialSpecification(
                    ApplyProcessorTask,
                    src_stack=self.src_stack,
                    processor_spec=this_proc,
                    pad=self.pad,
                    crop=self.crop,
                    )

            chunked_job = ChunkedJob(
                    task_class=this_task,
                    dst_layer=self.dst_layer,
                    chunk_xy=self.chunk_xy,
                    chunk_z=self.chunk_z,
                    blend_xy=self.blend_xy,
                    mip=this_proc_mip,
                    bcube=self.bcube
                    )


            yield from chunked_job.task_generator

            if not is_last_proc:
                yield scheduling.wait_until_done
                next_proc_mip = self.processor_mip[i + 1]
                if this_proc_mip > next_proc_mip:
                    downsample_job = DownsampleJob(
                                src_layer=self.dst_layer,
                                chunk_xy=self.chunk_xy,
                                chunk_z=self.chunk_z,
                                mip_start=this_proc_mip,
                                mip_end=next_proc_mip,
                                bcube=self.bcube
                                )
                    yield from downsample_job.task_generator
                    yield scheduling.wait_until_done


        if self.processor_mip[0] > self.processor_mip[-1]:
            # good manners
            # prepare the ground for the next you
            downsample_job = DownsampleJob(
                            src_layer=self.dst_layer,
                            chunk_xy=self.chunk_xy,
                            chunk_z=self.chunk_z,
                            mip_start=self.processor_mip[-1],
                            mip_end=self.processor_mip[0],
                            bcube=self.bcube
                            )


class ApplyProcessorTask(scheduling.Task):
    def __init__(self, processor_spec, src_stack, dst_layer,  mip,
            pad, crop, bcube):
        super().__init__()
        self.processor_spec = processor_spec
        self.src_stack = src_stack
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.crop = crop
        self.bcube = bcube

    def execute(self):
        src_bcube = self.bcube.uncrop(self.pad, self.mip)

        processor = procspec.parse_proc(
                spec_str=self.processor_spec)

        src_translation, src_data_dict = self.src_stack.read_data_dict(src_bcube,
                mip=self.mip, stack_name='src')

        processor_input = {**src_data_dict}

        result = processor(processor_input,
                           output_key='output')

        cropped_result = helpers.crop(result, self.crop)
        self.dst_layer.write(cropped_result, bcube=self.bcube, mip=self.mip)


@click.command()

@corgie_optgroup('Layer Parameters')

@corgie_option('--src_layer_spec',  '-s', nargs=1,
        type=str, required=True, multiple=True,
        help='Source layer spec. Use multiple times to include all masks, fields, images. ' + \
                LAYER_HELP_STR)
#
@corgie_option('--dst_layer_spec',  nargs=1,
        type=str, required=True,
        help= "Specification for the destination layer. Must be a field type." )

@corgie_option('--reference_key',        nargs=1, type=str, default='img')

@corgie_optgroup('Compute Field Method Specification')
@corgie_option('--chunk_xy',       '-c', nargs=1, type=int, default=1024)
@corgie_option('--chunk_z',              nargs=1, type=int, default=1)
@corgie_option('--blend_xy',             nargs=1, type=int, default=0)
@corgie_option('--pad',                  nargs=1, type=int, default=512,
        )
@corgie_optgroup('')
@corgie_option('--crop',                 nargs=1, type=int, default=None)
@corgie_option('--processor_spec',       nargs=1, type=str, multiple=True,
        required=True)
@corgie_option('--processor_mip',                  nargs=1, type=int, multiple=True,
        required=True)

@corgie_optgroup('Data Region Specification')
@corgie_option('--start_coord',      nargs=1, type=str, required=True)
@corgie_option('--end_coord',        nargs=1, type=str, required=True)
@corgie_option('--coord_mip',        nargs=1, type=int, default=0)


@click.pass_context
def apply_processor(ctx, src_layer_spec, dst_layer_spec,
        processor_spec, pad, crop, chunk_xy, start_coord, processor_mip,
        end_coord, coord_mip, blend_xy, chunk_z, reference_key):
    scheduler = ctx.obj['scheduler']

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec,
            name='src', readonly=True)

    reference_layer = None
    if reference_key in src_stack.layers:
        reference_layer = src_stack.layers[reference_key]

    dst_layer = create_layer_from_spec(dst_layer_spec, allowed_types=['img'],
            default_type='img', readonly=False, caller_name='dst_layer',
            reference=reference_layer, overwrite=True)

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if crop is None:
        crop = pad

    apply_processor_job = ApplyProcessorJob(
            src_stack=src_stack,
            dst_layer=dst_layer,
            chunk_xy=chunk_xy,
            chunk_z=chunk_z,
            blend_xy=blend_xy,
            processor_spec=processor_spec,
            pad=pad,
            crop=crop,
            bcube=bcube,
            processor_mip=processor_mip)

    # create scheduler and execute the job
    scheduler.register_job(apply_processor_job,
            job_name="Apply Processor {}".format(bcube))
    scheduler.execute_until_completion()



import os
import copy
import numpy as np

import cloudvolume as cv
from cloudvolume.lib import Bbox

from torch.nn.functional import interpolate

from corgie import layers
from corgie import exceptions
from corgie.log import logger as corgie_logger

from corgie.mipless_cloudvolume import MiplessCloudVolume
from corgie.data_backends.base import DataBackendBase, BaseLayerBackend, \
        register_backend

@register_backend("cv")
class CVDataBackend(DataBackendBase):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)


class CVLayerBase(BaseLayerBackend):
    def __init__(self, path, backend, reference=None, force_chunk_xy=None,
            force_chunk_z=None, overwrite=False, **kwargs):
        super().__init__(**kwargs)
        self.path = path
        self.backend = backend

        new_layer = False
        try:
            info = MiplessCloudVolume(path).get_info()
        except cv.exceptions.InfoUnavailableError as e:
            if reference is None:
                raise e
            else:
                info = copy.deepcopy(reference.get_info())
                new_layer = True

        if new_layer or overwrite:
            info['num_channels'] = self.get_num_channels()
            if self.dtype is None:
                dtype = self.get_default_data_type()
            else:
                dtype = self.dtype
            info['data_type'] = dtype
            if not self.supports_voxel_offset():
                for scale in info['scales']:
                    scale['voxel_offset'] = [0, 0, 0]
            if not self.supports_chunking():
                for scale in info['scales']:
                    scale['chunk_sizes'] = [[1, 1, 1]]
            if force_chunk_z is not None:
                for scale in info['scales']:
                    scale['chunk_sizes'][0][-1] = force_chunk_z
            if force_chunk_xy is not None:
                for scale in info['scales']:
                    scale['chunk_sizes'][0][0] = force_chunk_xy
                    scale['chunk_sizes'][0][1] = force_chunk_xy

        self.cv = MiplessCloudVolume(path,
                info=info, overwrite=overwrite)

        self.dtype = info['data_type']

    def __str__(self):
        return "CV {}".format(self.path)

    def get_sublayer(self, name, layer_type=None, path=None, **kwargs):
        if path is None:
            path = os.path.join(self.cv.path, layer_type, name)

        if layer_type is None:
            layer_type = self.get_layer_type()


        return self.backend.create_layer(path=path, reference=self,
                layer_type=layer_type, **kwargs)

    def read_backend(self, bcube, mip):
        x_range = bcube.x_range(mip)
        y_range = bcube.y_range(mip)
        z_range = bcube.z_range()

        this_cv = self.cv[mip]
        x_off, y_off, z_off = this_cv.voxel_offset
        '''if x_range[0] < x_off:
            corgie_logger.debug(f"READ from {str(self)}: \n"
                    f"   reducing xs from {x_range[0]} to {x_off} MIP: {mip}")
            x_range[0] = x_off
        if y_range[0] < y_off:
            corgie_logger.debug(f"READ from {str(self)}: \n"
                    f"   reducing ys from {y_range[0]} to {y_off} MIP: {mip}")
            y_range[0] = y_off
        if z_range[0] < z_off:
            corgie_logger.debug(f"READ from {str(self)}: \n"
                    f"   reducing zs from {z_range[0]} to {z_off} MIP: {mip}")
            z_range[0] = z_off'''

        corgie_logger.debug("READ from {}: \n   x: {}, y: {}, z: {}, MIP: {}".format(
            str(self), x_range, y_range, z_range, mip))
        data = self.cv[mip][x_range[0]:x_range[1],
                     y_range[0]:y_range[1],
                     z_range[0]:z_range[1]]
        data = np.transpose(data, (2,3,0,1))
        return data

    def write_backend(self, data, bcube, mip):
        x_range = bcube.x_range(mip)
        y_range = bcube.y_range(mip)
        z_range = bcube.z_range()

        data = np.transpose(data, (2,3,0,1))
        corgie_logger.debug("Write to {}: \n x: {}, y: {}, z: {}, MIP: {}".format(
            str(self), x_range, y_range, z_range, mip))
        self.cv[mip].autocrop = True
        self.cv[mip][x_range[0]:x_range[1],
                     y_range[0]:y_range[1],
                     z_range[0]:z_range[1]] = data
        self.cv[mip].autocrop = False


    def get_info(self):
        return self.cv.get_info()

    def get_chunk_aligned_bcube(self, bcube, mip, chunk_xy, chunk_z):
        cv_chunk = self.cv[mip].chunk_size

        # Expand bbox to be difizible by chunk_size

        bbox = Bbox((bcube.x_range(mip)[0], bcube.y_range(mip)[0], bcube.z_range()[0]),
                    (bcube.x_range(mip)[1], bcube.y_range(mip)[1], bcube.z_range()[1]))

        aligned_bbox = bbox.expand_to_chunk_size(self.cv[mip].chunk_size,
                                                 self.cv[mip].voxel_offset)

        aligned_bcube = copy.deepcopy(bcube)
        aligned_bcube.reset_coords(aligned_bbox.minpt[0], aligned_bbox.maxpt[0],
                                   aligned_bbox.minpt[1], aligned_bbox.maxpt[1],
                                   aligned_bbox.minpt[2], aligned_bbox.maxpt[2],
                                   mip=mip)
        if chunk_xy is not None:

            if chunk_xy % cv_chunk[0] != 0:
                raise exceptions.ChunkingError(self, f"Processing chunk_xy {chunk_xy} is not"
                        f"divisible by MIP{mip} CV chunk {cv_chunk[0]}")
            if chunk_z % cv_chunk[2] != 0:
                raise exceptions.ChunkingError(self, f"Processing chunk_z {chunk_z} is not"
                        f"divisible by MIP{mip} CV chunk {cv_chunk[2]}")

            if chunk_xy > aligned_bcube.x_size(mip):
                x_adj = chunk_xy - aligned_bcube.x_size(mip)
            else:
                x_rem = aligned_bcube.x_size(mip) % chunk_xy
                if x_rem == 0:
                    x_adj = 0
                else:
                    x_adj = chunk_xy - x_rem
            if chunk_xy > aligned_bcube.y_size(mip):
                y_adj = chunk_xy - aligned_bcube.y_size(mip)
            else:
                y_rem = aligned_bcube.y_size(mip) % chunk_xy
                if y_rem == 0:
                    y_adj = 0
                else:
                    y_adj = chunk_xy - y_rem

            if chunk_z > aligned_bcube.z_size():
                z_adj = chunk_z - aligned_bcube.z_size()
            else:
                rem = aligned_bcube.z_size() % chunk_z
                if rem == 0:
                    z_adj = 0
                else:
                    z_adj = chunk_z - z_rem
            if x_adj != 0:
                xe = aligned_bcube.x_range(mip)[1]
                aligned_bcube.reset_coords(xe=xe + x_adj, mip=mip)
            if y_adj != 0:
                ye = aligned_bcube.y_range(mip)[1]
                aligned_bcube.reset_coords(ye=ye + y_adj, mip=mip)
            if z_adj != 0:
                ze = aligned_bcube.z_range()[1]
                aligned_bcube.reset_coords(ze=ze + z_adj)
        return aligned_bcube

    def break_bcube_into_chunks(self, bcube, chunk_xy, chunk_z, mip,
            readonly=False, **kwargs):
        if not readonly:
            bcube = self.get_chunk_aligned_bcube(bcube, mip, chunk_xy, chunk_z)

        chunks = super().break_bcube_into_chunks(bcube, chunk_xy, chunk_z,
                mip, **kwargs)
        return chunks


@CVDataBackend.register_layer_type_backend("img")
class CVImgLayer(CVLayerBase, layers.ImgLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)


@CVDataBackend.register_layer_type_backend("field")
class CVFieldLayer(CVLayerBase, layers.FieldLayer):
    supported_backend_dtypes = ['float32', 'int16']
    def __init__(self, backend_dtype='float32', **kwargs):
        super().__init__(**kwargs)
        if backend_dtype not in self.supported_backend_dtypes:
            raise exceptions.ArgumentError("Field layer 'backend_type'",
                    "\n{} is not a supported field backend data type. \n"
                    "Supported backend data types: {}".format(backend_type,
                        self.supported_backend_dtypes)
                    )
        self.backend_dtype = backend_dtype


@CVDataBackend.register_layer_type_backend("mask")
class CVMaskLayer(CVLayerBase, layers.MaskLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)


@CVDataBackend.register_layer_type_backend("section_value")
class CVSectionValueLayer(CVLayerBase, layers.SectionValueLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)



import torch

def shift_by_int(img, x_shift, y_shift, is_res=False):
    if is_res:
        img = img.permute(0, 3, 1, 2)

    x_shifted = torch.zeros_like(img)
    if x_shift > 0:
        x_shifted[..., x_shift:, :]  = img[..., :-x_shift, :]
    elif x_shift < 0:
        x_shifted[..., :x_shift, :]  = img[..., -x_shift:, :]
    else:
        x_shifted = img.clone()

    result = torch.zeros_like(img)
    if y_shift > 0:
        result[..., y_shift:]  = x_shifted[..., :-y_shift]
    elif y_shift < 0:
        result[..., :y_shift]  = x_shifted[..., -y_shift:]
    else:
        result = x_shifted.clone()

    if is_res:
        result = result.permute(0, 2, 3, 1)

    return result


def res_warp_res(res_a, res_b, is_pix_res=True, permute_field=True):
    if is_pix_res:
        res_b = 2 * res_b / (res_b.shape[-2])

    if len(res_a.shape) == 4:
        result = gridsample_residual(
                        res_a.permute(0, 3, 1, 2),
                        res_b,
                        padding_mode='border').permute(0, 2, 3, 1)
    elif len(res_a.shape) == 3:
        result = gridsample_residual(
                        res_a.permute(2, 0, 1).unsqueeze(0),
                        res_b.unsqueeze(0),
                        padding_mode='border')[0].permute(1, 2, 0)
    else:
        raise Exception("Residual warping requires BxHxWx2 or HxWx2 format.")

    return result


def res_warp_img(img, res_in, is_pix_res=True,
        padding_mode='zeros', mode="bilinear",
        permute_field=True):
    if permute_field:
        res_in = res_in.permute(0, 2, 3, 1)

    if is_pix_res:
        res = 2 * res_in / (img.shape[-1])
    else:
        res = res_in
    if len(img.shape) == 4:
        result = gridsample_residual(img, res, padding_mode=padding_mode, mode=mode)
    elif len(img.shape) == 3:
        if len(res.shape) == 3:
            result = gridsample_residual(img.unsqueeze(0),
                                         res.unsqueeze(0), padding_mode=padding_mode,
                                         mode=mode)[0]
        else:
            img = img.unsqueeze(1)
            result = gridsample_residual(img,
                                         res,
                                         mode=mode,
                                         padding_mode=padding_mode).squeeze(1)
    elif len(img.shape) == 2:
        result = gridsample_residual(img.unsqueeze(0).unsqueeze(0),
                                     res.unsqueeze(0),
                                     padding_mode=padding_mode,
                                     mode=mode)[0, 0]
    else:
        raise Exception("Image warping requires BxCxHxW or CxHxW format." +
                        "Recieved dimensions: {}".format(len(img.shape)))

    return result


def combine_residuals(a, b, is_pix_res=True):
    return b + res_warp_res(a, b, is_pix_res=is_pix_res)


def upsample_residuals(residuals, factor=2.0):
    original_dim = len(residuals.shape)
    while len(residuals.shape) < 4:
        residuals = residuals.unsqueeze(0)
    res_perm = residuals.permute(0, 3, 1, 2)
    result = torch.nn.functional.interpolate(res_perm, scale_factor=factor, mode='bicubic').permute(0, 2, 3, 1)
    result *= factor
    while len(result.shape) > original_dim:
        result = result.squeeze(0)
    return result


def downsample_residuals(residuals):
    original_dim = len(residuals.shape)
    while len(residuals.shape) < 4:
        residuals = residuals.unsqueeze(0)
    result = torch.nn.functional.avg_pool2d(residuals.permute(
                                     0, 3, 1, 2), 2).permute(0, 2, 3, 1)
    result /= 2
    while len(result.shape) > original_dim:
        result = result.squeeze(0)
    return result


def gridsample(source, field, padding_mode, mode='bilinear'):
    """
    A version of the PyTorch grid sampler that uses size-agnostic conventions.
    Vectors with values -1 or +1 point to the actual edges of the images
    (as opposed to the centers of the border pixels as in PyTorch 4.1).
    `source` and `field` should be PyTorch tensors on the same GPU, with
    `source` arranged as a PyTorch image, and `field` as a PyTorch vector
    field.
    `padding_mode` is required because it is a significant consideration.
    It determines the value sampled when a vector is outside the range [-1,1]
    Options are:
     - "zero" : produce the value zero (okay for sampling images with zero as
                background, but potentially problematic for sampling masks and
                terrible for sampling from other vector fields)
     - "border" : produces the value at the nearest inbounds pixel (great for
                  masks and residual fields)
    If sampling a field (ie. `source` is a vector field), best practice is to
    subtract out the identity field from `source` first (if present) to get a
    residual field.
    Then sample it with `padding_mode = "border"`.
    This should behave as if source was extended as a uniform vector field
    beyond each of its boundaries.
    Note that to sample from a field, the source field must be rearranged to
    fit the conventions for image dimensions in PyTorch. This can be done by
    calling `source.permute(0,3,1,2)` before passing to `gridsample()` and
    `result.permute(0,2,3,1)` to restore the result.
    """
    if source.shape[2] != source.shape[3]:
        raise NotImplementedError('Grid sampling from non-square tensors '
                                  'not yet implementd here.')
    scaled_field = field * source.shape[2] / (source.shape[2] - 1)
    return torch.nn.functional.grid_sample(source, scaled_field, mode=mode,
                         padding_mode=padding_mode)


def gridsample_residual_2d(source, residual, padding_mode):
    """
    Similar to `gridsample()`, but takes a residual field.
    This abstracts away generation of the appropriate identity grid.
    """
    source = torch.FloatTensor(source).unsqueeze(0).unsqueeze(0)
    residual = torch.FloatTensor(residual).unsqueeze(0)
    return gridsample_residual(source, residual, padding_mode)


def gridsample_residual(source, residual, padding_mode, mode="bilinear"):
    """
    Similar to `gridsample()`, but takes a residual field.
    This abstracts away generation of the appropriate identity grid.
    """
    field = residual + identity_grid(residual.shape, device=residual.device)
    return gridsample(source, field, padding_mode, mode=mode)


def _create_identity_grid(size, device='cpu'):
    with torch.no_grad():
        id_theta = torch.FloatTensor([[[1,0,0],[0,1,0]]]).to(device) # identity affine transform
        I = torch.nn.functional.affine_grid(id_theta,torch.Size((1,1,size,size)))
        I *= (size - 1) / size # rescale the identity provided by PyTorch
        return I


def identity_grid(size, cache=False, device='cpu'):
    """
    Returns a size-agnostic identity field with -1 and +1 pointing to the
    corners of the image (not the centers of the border pixels as in
    PyTorch 4.1).
    Use `cache = True` to cache the identity for faster recall.
    This can speed up recall, but may be a burden on cpu/gpu memory.
    `size` can be either an `int` or a `torch.Size` of the form
    `(N, C, H, W)`. `H` and `W` must be the same (a square tensor).
    `N` and `C` are ignored.
    """
    if isinstance(size, torch.Size):
        if (size[2] == size[3] # image
            or (size[3] == 2 and size[1] == size[2])): # field
            size = size[2]
        else:
            raise ValueError("Bad size: {}. Expected a square tensor size.".format(size))
    if size in identity_grid._identities:
        return identity_grid._identities[size].to(device)
    I = _create_identity_grid(size, device)
    if cache:
        identity_grid._identities[size] = I
    return I.to(device)
identity_grid._identities = {}



import click

from corgie import scheduling, helpers, stack
from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option, \
        create_stack_from_spec

class CopyJob(scheduling.Job):
    def __init__(self, src_stack, dst_stack, mip, copy_masks,
                 blackout_masks, bcube, chunk_xy, chunk_z):
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.mip = mip
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.copy_masks = copy_masks
        self.blackout_masks = blackout_masks

        super().__init__()

    def task_generator(self):
        chunks = self.dst_stack.get_layers()[0].break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip=self.mip)

        tasks = [CopyTask(self.src_stack,
                            self.dst_stack,
                            blackout_masks=self.blackout_masks,
                            copy_masks=self.copy_masks,
                            mip=self.mip,
                            bcube=input_chunk) for input_chunk in chunks]
        corgie_logger.info(f"Yielding copy tasks for bcube: {self.bcube}, MIP: {self.mip}")

        yield tasks


class CopyTask(scheduling.Task):
    def __init__(self, src_stack, dst_stack, copy_masks, blackout_masks, mip,
                 bcube):
        super().__init__(self)
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.copy_masks = copy_masks
        self.blackout_masks = blackout_masks
        self.mip = mip
        self.bcube = bcube

    def execute(self):
        src_translation, src_data_dict = self.src_stack.read_data_dict(self.bcube,
                mip=self.mip, translation_adjuster=None, stack_name='src')

        if self.blackout_masks:
            mask_layers = self.dst_stack.get_layers_of_type(["mask"])
            mask = helpers.read_mask_list(mask_layers, self.bcube, self.mip)
        else:
            mask = None
        if self.copy_masks:
            write_layers = self.dst_stack.get_layers_of_type(["img", "mask"])
        else:
            write_layers = self.dst_stack.get_layers_of_type("img")

        for l in write_layers:
            src = src_data_dict[f"src_{l.name}"]
            if mask is not None:
                src[mask] = 0
            l.write(src, bcube=self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup('Layer Parameters')
@corgie_option('--src_layer_spec',  '-s', nargs=1,
        type=str, required=True, multiple=True,
        help='Source layer spec. Use multiple times to include all masks, fields, images. ' + \
                LAYER_HELP_STR)
#
@corgie_option('--dst_folder',  nargs=1,
        type=str, required=False,
        help= "Destination folder for the copied stack")

@corgie_optgroup('Copy Method Specification')
@corgie_option('--chunk_xy',       '-c', nargs=1, type=int, default=1024)
@corgie_option('--chunk_z',              nargs=1, type=int, default=1)
@corgie_option('--mip',                  nargs=1, type=int, required=True)
@corgie_option('--blackout_masks/--no_blackout_masks',  default=False)
@corgie_option('--copy_masks/--no_copy_masks',          default=True)
@corgie_option('--force_chunk_xy',  is_flag=True)
@corgie_option('--force_chunk_z',      is_flag=True)

@corgie_optgroup('Data Region Specification')
@corgie_option('--start_coord',      nargs=1, type=str, required=True)
@corgie_option('--end_coord',        nargs=1, type=str, required=True)
@corgie_option('--coord_mip',        nargs=1, type=int, default=0)
@corgie_option('--suffix',           nargs=1, type=str, default=None)

@click.pass_context
def copy(ctx, src_layer_spec, dst_folder, copy_masks, blackout_masks,
         chunk_xy, chunk_z, start_coord, end_coord, coord_mip, mip, suffix,
         force_chunk_xy, force_chunk_z):

    scheduler = ctx.obj['scheduler']
    if suffix is None:
        suffix = ''
    else:
        suffix = f"_{suffix}"

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec,
            name='src', readonly=True)

    if force_chunk_xy:
        force_chunk_xy = chunk_xy
    else:
        force_chunk_xy = None

    if force_chunk_z:
        force_chunk_z = chunk_z
    else:
        force_chunk_z = None

    dst_stack = stack.create_stack_from_reference(reference_stack=src_stack,
            folder=dst_folder, name="dst", types=["img", "mask"], readonly=False,
            suffix=suffix, force_chunk_xy=force_chunk_xy, force_chunk_z=force_chunk_z, overwrite=True)

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    copy_job = CopyJob(src_stack=src_stack,
                           dst_stack=dst_stack,
                           mip=mip,
                           bcube=bcube,
                           chunk_xy=chunk_xy,
                           chunk_z=chunk_z,
                           copy_masks=copy_masks,
                           blackout_masks=blackout_masks)

    # create scheduler and execute the job
    scheduler.register_job(copy_job, job_name="Copy {}".format(bcube))
    scheduler.execute_until_completion()



import os
import click
from copy import deepcopy

from corgie import scheduling
from corgie import helpers, stack
from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie import argparsers
from corgie.cli.compute_stats import ComputeStatsJob
from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option, \
        create_stack_from_spec

class NormalizeJob(scheduling.Job):
    def __init__(self, src_layer, mask_layers, dst_layer, mean_layer, var_layer, stats_mip,
            mip, bcube, chunk_xy, chunk_z, mask_value):
        self.src_layer = src_layer
        self.mask_layers = mask_layers
        self.dst_layer = dst_layer
        self.var_layer = var_layer
        self.mean_layer = mean_layer
        self.stats_mip = stats_mip
        self.mip = mip
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.mask_value = mask_value
        super().__init__()

    def task_generator(self):
        chunks = self.dst_layer.break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip=self.mip)

        tasks = [NormalizeTask(self.src_layer,
                                self.mask_layers,
                                self.dst_layer,
                                self.mean_layer,
                                self.var_layer,
                                self.stats_mip,
                                self.mip,
                                self.mask_value,
                                input_chunk) for input_chunk in chunks]
        print (f"Yielding {len(tasks)} normalize tasks for bcube: {self.bcube}, MIP: {self.mip}")

        yield tasks


class NormalizeTask(scheduling.Task):
    def __init__(self, src_layer, mask_layers, dst_layer, mean_layer, var_layer,
            stats_mip, mip, mask_value, bcube):
        super().__init__()
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mean_layer = mean_layer
        self.mask_layers = mask_layers
        self.var_layer = var_layer
        self.stats_mip = stats_mip
        self.mip = mip
        self.bcube = bcube
        self.mask_value = mask_value

    def execute(self):
        corgie_logger.info(f"Normalizing {self.src_layer} at MIP{self.mip}, region: {self.bcube}")
        mean_data = self.mean_layer.read(self.bcube, mip=self.stats_mip)
        var_data = self.var_layer.read(self.bcube, mip=self.stats_mip)

        src_data = self.src_layer.read(self.bcube, mip=self.mip)
        mask_data = helpers.read_mask_list(
                mask_list=self.mask_layers,
                bcube=self.bcube, mip=self.mip)

        dst_data = (src_data - mean_data) / var_data.sqrt()
        if mask_data is not None:
            dst_data[mask_data] = self.mask_value
        self.dst_layer.write(dst_data, self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup('Layer Parameters')

@corgie_option('--src_layer_spec',  '-s', nargs=1,
        type=str, required=True, multiple=True,
        help='Source layer spec. Use multiple times to include all masks, fields, images. ' + \
                LAYER_HELP_STR)

@corgie_option('--dst_folder',  nargs=1, type=str, required=True,
        help="Folder where aligned stack will go")

@corgie_option('--suffix',     '-s', nargs=1, type=str, default=None)

@corgie_optgroup('Normalize parameters')
@corgie_option('--recompute_stats/--no_recompute_stats',  default=False )
@corgie_option('--stats_mip',  '-m', nargs=1, type=int, required=None)
@corgie_option('--mip_start',  '-m', nargs=1, type=int, required=True)
@corgie_option('--mip_end',    '-e', nargs=1, type=int, required=True)
@corgie_option('--chunk_xy',   '-c', nargs=1, type=int, default=2048)
@corgie_option('--chunk_z',          nargs=1, type=int, default=1)
@corgie_option('--mask_value',       nargs=1, type=float, default=0.0)

@corgie_optgroup('Data Region Specification')
@corgie_option('--start_coord',      nargs=1, type=str, required=True)
@corgie_option('--end_coord',        nargs=1, type=str, required=True)
@corgie_option('--coord_mip',        nargs=1, type=int, default=0)
@click.pass_context
def normalize(ctx, src_layer_spec, dst_folder, stats_mip,
        mip_start, mip_end, chunk_xy, chunk_z, start_coord,
        end_coord, coord_mip, suffix, recompute_stats, mask_value):
    if chunk_z != 1:
        raise NotImplemented("Compute Statistics command currently only \
                supports per-section statistics.")
    result_report = ""
    scheduler = ctx.obj['scheduler']

    if suffix is None:
        suffix = '_norm'
    else:
        suffix = f"_{suffix}"

    if stats_mip is None:
        stats_mip = mip_end

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec,
            name='src', readonly=True)

    dst_stack = stack.create_stack_from_reference(reference_stack=src_stack,
            folder=dst_folder, name="dst", types=["img"], readonly=False,
            suffix=suffix, overwrite=True)

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    img_layers = src_stack.get_layers_of_type("img")
    mask_layers = src_stack.get_layers_of_type("mask")
    field_layers = src_stack.get_layers_of_type("field")
    assert len(field_layers) == 0

    for l in img_layers:
        mean_layer = l.get_sublayer(
                    name=f"mean{suffix}",
                    path=os.path.join(dst_folder, f"mean{suffix}"),
                    layer_type="section_value",
                    )

        var_layer  = l.get_sublayer(
                name=f"var{suffix}",
                path=os.path.join(dst_folder, f"var{suffix}"),
                layer_type="section_value")

        if recompute_stats:
            compute_stats_job = ComputeStatsJob(
                   src_layer=l,
                   mask_layers=mask_layers,
                   mean_layer=mean_layer,
                   var_layer=var_layer,
                   bcube=bcube,
                   mip=stats_mip,
                   chunk_xy=chunk_xy,
                   chunk_z=chunk_z)

        # create scheduler and execute the job
            scheduler.register_job(compute_stats_job, job_name=f"Compute Stats. Layer: {l}, Bcube: {bcube}")
            scheduler.execute_until_completion()

        dst_layer = l.get_sublayer(
                    name=f"{l.name}{suffix}",
                    path=os.path.join(dst_folder, "img", f"{l.name}{suffix}"),
                    layer_type=l.get_layer_type(),
                    dtype='float32',
                    overwrite=True
                    )

        result_report += f"Normalized {l} -> {dst_layer}\n"
        for mip in range(mip_start, mip_end + 1):
            normalize_job = NormalizeJob(src_layer=l,
                    mask_layers=mask_layers,
                    dst_layer=deepcopy(dst_layer),
                    mean_layer=mean_layer,
                    var_layer=var_layer,
                    stats_mip=stats_mip,
                    mip=mip,
                    bcube=bcube,
                    chunk_xy=chunk_xy,
                    chunk_z=chunk_z,
                    mask_value=mask_value)

            # create scheduler and execute the job
            scheduler.register_job(normalize_job, job_name=f"Normalize {bcube}, MIP {mip}")
    scheduler.execute_until_completion()
    corgie_logger.info(result_report)



import logging
import sys

logger = logging.getLogger('corgie')


def configure_logger(verbose):
    log_level = logging.DEBUG if verbose else logging.INFO
    logger.setLevel(log_level)
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    logger.addHandler(ch)



import json
import six
import copy

import cloudvolume
from cloudvolume import CloudVolume, Storage

from corgie.log import logger as corgie_logger

def deserialize_miplessCV_old(s, cache={}):
        if s in cache:
            return cache[s]
        else:
            contents = json.loads(s)
            mcv = MiplessCloudVolume(contents['path'], mkdir=contents['mkdir'],
                                                             **contents['kwargs'])
            cache[s] = mcv
            return mcv

def deserialize_miplessCV_old2(s, cache={}):
        cv_kwargs = {'bounded': False, 'progress': False,
                     'autocrop': False, 'non_aligned_writes': False,
                     'cdn_cache': False}
        if s in cache:
            return cache[s]
        else:
            contents = json.loads(s)
            mcv = MiplessCloudVolume(contents['path'], mkdir=False,
                                                             fill_missing=True, **cv_kwargs)
            cache[s] = mcv
            return mcv

def deserialize_miplessCV(s, cache={}):
        cv_kwargs = {'bounded': False, 'progress': False,
                    'autocrop': False, 'non_aligned_writes': False,
                    'cdn_cache': False}
        if s in cache:
            return cache[s]
        else:
            mcv = MiplessCloudVolume(s, mkdir=False,
                                                             fill_missing=True, **cv_kwargs)
            cache[s] = mcv
            return mcv

class MiplessCloudVolume():
    """Multi-mip access to CloudVolumes using the same path
    """
    def __init__(self, path, allow_info_writes=True, obj=CloudVolume,
            default_chunk=(512, 512, 1), overwrite=False, **kwargs):
        self.path = path
        self.allow_info_writes = allow_info_writes
        self.cv_params = {}
        self.cv_params['bounded'] = False
        self.cv_params['progress'] = False
        self.cv_params['autocrop'] = False
        self.cv_params['non_aligned_writes'] = False
        self.cv_params['cdn_cache'] = False
        self.cv_params['fill_missing'] = True

        for k, v in six.iteritems(kwargs):
            self.cv_params[k] = v

        self.default_chunk = default_chunk

        self.obj = obj
        self.cvs = {}
        if 'info' in self.cv_params and overwrite:
            self.store_info()

    # def exists(self):
    #       s = Storage(self.path)
    #       return s.exists('info')

    def serialize(self):
            contents = {
                   "path" : self.path,
                   "allow_info_writes" : self.allow_info_writes,
                   "cv_params": self.cv_params,
            }
            s = json.dumps(contents)
            return s

    @classmethod
    def deserialize(cls, s, cache={}, **kwargs):
            if s in cache:
                return cache[s]
            else:
                import pdb; pdb.set_trace()
                mcv = cls(s,  **kwargs)
                cache[s] = mcv
                return mcv

    def get_info(self):
        tmp_cv = self.obj(self.path, **self.cv_params)
        return tmp_cv.info

    def store_info(self, info=None):
        if not self.allow_info_writes:
            raise Exception("Attempting to store info to {}, but "
                    "'allow_info_writes' flag is set to False".format(self.path))

        tmp_cv = self.obj(self.path, **self.cv_params)
        if info is not None:
            tmp_cv.info = info

        tmp_cv.commit_info()
        tmp_cv.commit_provenance()

    def ensure_info_has_mip(self, mip):
        tmp_cv = self.obj(self.path, **self.cv_params)
        scale_num = len(tmp_cv.info['scales'])

        if scale_num < mip + 1:
            while scale_num < mip + 1:
                tmp_cv.add_scale((2**scale_num, 2**scale_num, 1),
                        chunk_size=self.default_chunk)
                scale_num += 1

            self.store_info(tmp_cv.info)

    def extend_info_to_mip(self, mip):
        info = self.get_info()
        highest_mip = len(info['scales']) - 1
        highest_mip_info = info['scales'][-1]

        if highest_mip >= mip:
            return

        while (highest_mip < mip):
            new_highest_mip_info = copy.deepcopy(highest_mip_info)
            #size
            #voxel offset
            #resolution -> key
            new_highest_mip_info['size'] = [
                    highest_mip_info['size'][0] // 2,
                    highest_mip_info['size'][1] // 2,
                    highest_mip_info['size'][2]
                ]

            new_highest_mip_info['voxel_offset'] = [
                    highest_mip_info['voxel_offset'][0] // 2,
                    highest_mip_info['voxel_offset'][1] // 2,
                    highest_mip_info['voxel_offset'][2]
                ]

            new_highest_mip_info['resolution'] = [
                    highest_mip_info['resolution'][0] * 2,
                    highest_mip_info['resolution'][1] * 2,
                    highest_mip_info['resolution'][2]
                ]

            new_highest_mip_info['key'] = '_'.join([str(i) for i in new_highest_mip_info['resolution']])
            info['scales'].append(new_highest_mip_info)
            highest_mip += 1
            highest_mip_info = new_highest_mip_info

        self.store_info()

    def create(self, mip):

        corgie_logger.debug('Creating CloudVolume for {0} at MIP{1}'.format(self.path, mip))
        self.extend_info_to_mip(mip)

        self.cvs[mip] = self.obj(self.path, mip=mip, **self.cv_params)


        #if self.mkdir:
        #  self.cvs[mip].commit_info()
        #  self.cvs[mip].commit_provenance()

    def __getitem__(self, mip):
        if mip not in self.cvs:
            self.create(mip)
        return self.cvs[mip]

    def __repr__(self):
        return self.path




import math
import torch

import click
from click_option_group import optgroup

from corgie import scheduling
from corgie import helpers
from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie import argparsers
from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option


class ComputeStatsJob(scheduling.Job):
    def __init__(self, src_layer, mask_layers, mean_layer, var_layer, bcube,
            mip, chunk_xy, chunk_z):
        self.src_layer = src_layer
        self.mask_layers = mask_layers
        self.mean_layer = mean_layer
        self.var_layer = var_layer
        self.mip = mip
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z

        super().__init__(self)

    def task_generator(self):
        chunks = self.src_layer.break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip=self.mip)

        assert len(chunks) % self.bcube.z_size() == 0
        chunks_per_section = len(chunks) // self.bcube.z_size()

        chunk_mean_layer  = self.mean_layer.get_sublayer(
                name="chunk_mean",
                layer_type="section_value",
                num_channels=chunks_per_section)

        chunk_var_layer  = self.var_layer.get_sublayer(
                name="chunk_var",
                layer_type="section_value",
                num_channels=chunks_per_section)

        chunks = self.src_layer.break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip=self.mip)

        # sort chunks by z
        chunks.sort(reverse=True, key=lambda c:c.z_range()[1])

        tasks = [ComputeStatsTask(self.src_layer,
                                mean_layer=chunk_mean_layer,
                                mask_layers=self.mask_layers,
                                var_layer=chunk_var_layer,
                                mip=self.mip,
                                bcube=chunks[chunk_num],
                                # chunks are sorted by z, so this gives the chunk num
                                # for a given z
                                write_channel=chunk_num % chunks_per_section) \
                            for chunk_num in range(len(chunks))]

        corgie_logger.info("Yielding chunk stats tasks: {}, MIP: {}".format(
            self.bcube, self.mip))
        yield tasks
        yield scheduling.wait_until_done

        accum_chunks = chunk_mean_layer.break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=chunks_per_section,
                chunk_z=self.chunk_z,
                mip=self.mip)

        accum_mean_tasks = [ComputeStatsTask(chunk_mean_layer,
                                mean_layer=self.mean_layer,
                                var_layer=None,
                                mip=self.mip,
                                bcube=accum_chunk,
                                write_channel=0) \
                            for accum_chunk in accum_chunks]

        accum_var_tasks = [ComputeStatsTask(chunk_var_layer,
                                mean_layer=self.var_layer,
                                var_layer=None,
                                mip=self.mip,
                                bcube=accum_chunk,
                                write_channel=0) \
                            for accum_chunk in accum_chunks]

        corgie_logger.info("Yielding chunk stats aggregation tasks...")
        yield accum_mean_tasks + accum_var_tasks

    def create_dst_layers(self):
        return mean_layer, var_layer


class ComputeStatsTask(scheduling.Task):
    def __init__(self, src_layer, mean_layer, var_layer, mip,
                 bcube, write_channel, mask_layers=[]):
        super().__init__(self)
        self.src_layer = src_layer
        self.mask_layers = mask_layers
        self.mean_layer = mean_layer
        self.var_layer = var_layer
        self.mip = mip
        self.bcube = bcube
        self.write_channel = write_channel

    def execute(self):
        src_data = self.src_layer.read(bcube=self.bcube,
                mip=self.mip)
        mask_data = helpers.read_mask_list(
                mask_list=self.mask_layers,
                bcube=self.bcube, mip=self.mip)

        if mask_data is not None:
            src_data = src_data[mask_data == 0]
        src_data = src_data[src_data != 0]
        src_data = src_data[src_data == src_data]

        if self.mean_layer is not None:
            if len(src_data) == 0:
                mean = torch.FloatTensor([0], device=src_data.device)
            else:
                mean = src_data.float().mean()
            self.mean_layer.write(
                    mean,
                    bcube=self.bcube,
                    mip=self.mip,
                    channel_start=self.write_channel,
                    channel_end=self.write_channel + 1)

        if self.var_layer is not None:
            if len(src_data) == 0:
                var = torch.FloatTensor([0], device=src_data.device)
            else:
                var = src_data.float().var()

            self.var_layer.write(
                    var,
                    bcube=self.bcube,
                    mip=self.mip,
                    channel_start=self.write_channel,
                    channel_end=self.write_channel + 1)


@click.command()
@corgie_optgroup('Layer parameters')
@corgie_option('--src_layer_spec',  '-s', nargs=1,
        type=str, required=True, multiple=True,
        help='Source layer spec. Use multiple times to include all masks, fields, images. ' + \
                LAYER_HELP_STR)

@corgie_option('--dst_folder',  nargs=1, type=str, required=True,
        help="Folder where aligned stack will go")


# Other Params
@corgie_option('--suffix',     '-s', nargs=1, type=str, default=None)

@corgie_optgroup('Computation Specification')
@corgie_option('--mip',        '-m', nargs=1, type=int, required=True)
@corgie_option('--chunk_xy',   '-c', nargs=1, type=int, default=4096)
@corgie_option('--chunk_z',          nargs=1, type=int, default=1)

@corgie_optgroup('Data Region Specification')
@corgie_option('--start_coord',      nargs=1, type=str, required=True)
@corgie_option('--end_coord',        nargs=1, type=str, required=True)
@corgie_option('--coord_mip',        nargs=1, type=int, default=0)
@click.pass_context
def compute_stats(ctx, src_layer_spec, dst_folder, suffix, mip,
        chunk_xy, chunk_z,  start_coord, end_coord, coord_mip):
    compute_stats_fn(ctx, src_layer_spec, mask_layers_spec, suffix, mip,
        chunk_xy, chunk_z,  start_coord, end_coord, coord_mip)


def compute_stats_fn(ctx, src_layer_spec, dst_folder, suffix, mip,
        chunk_xy, chunk_z,  start_coord, end_coord, coord_mip):

    if chunk_z != 1:
        raise NotImplemented("Compute Statistics command currently only \
                supports per-section statistics.")

    scheduler = ctx.obj['scheduler']

    src_layer = create_layer_from_spec(src_layer_spec,
            caller_name='src layer',
            readonly=True)

    mask_layers = [create_layer_from_spec(mask_spec,
            caller_name='src mask',
            readonly=True,
            allowed_types=['mask'],
            default_type='mask') for mask_spec in mask_layers_spec
    ]

    if suffix is None:
        suffix = ''
    else:
        suffix = '_' + suffix

    src_stack = create_stack_from_spec(src_layer_spec,
            name='src', readonly=True)

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    mask_layers = src_stack.get_layers_of_type(["mask"])
    non_mask_layers = src_stack.get_layers_of_type(["img", "field"])

    for l in non_mask_layers:
        mean_layer = src_layer.get_sublayer(
                name=f"mean{suffix}",
                path=os.path.join(dst_dir, f"mean{suffix}"),
                layer_type="section_value",
                )

        var_layer  = src_layer.get_sublayer(
                path=os.path.join(dst_dir, f"var{suffix}"),
                layer_type="section_value")


        compute_stats_job = ComputeStatsJob(
               src_layer=l,
               mask_layers=mask_layers,
               mean_layer=mean_layer,
               var_layer=var_layer,
               bcube=bcube,
               mip=mip,
               chunk_xy=chunk_xy,
               chunk_z=chunk_z)

        # create scheduler and execute the job
        scheduler.register_job(compute_stats_job, job_name=f"Compute Stats. Layer: {l}, Bcube: {bcube}")
    scheduler.execute_until_completion()






class CorgieException(Exception):
    pass


class IncorrectArgumentDefinition(CorgieException):
    def __init__(self, command, argument, argtype, reason=''):
        self.msg = "\nCommand: {}, Argument: {} {}\n Reason: {}".format(command,
                argtype, argument, reason)
        super().__init__(self.msg)


class ArgumentError(CorgieException):
    def __init__(self, argument, reason=''):
        self.msg = "\nArgument: {}\n Reason: {}".format(argument, reason)
        super().__init__(self.msg)


class ReadError(CorgieException):
    def __init__(self, layer, reason):
        self.msg = "\nLayer '{}': \n{}".format(str(layer), reason)
        super().__init__(self.msg)


class ChunkingError(CorgieException):
    def __init__(self, layer, reason):
        self.msg = "\nLayer '{}': \n{}".format(str(layer), reason)
        super().__init__(self.msg)


class WriteError(CorgieException):
    def __init__(self, layer, reason):
        self.msg = "\nLayer '{}': \n{}".format(str(layer), reason)
        super().__init__(self.msg)



# Stack class
# Stack of sections
# defined by:
# 3d bounding box
# cv
# section is a stack with thickness == 1
import six
import os
import copy

from corgie import exceptions, helpers
from corgie.layers import str_to_layer_type

class StackBase:
    def __init__(self, name=None):
        self.name = name
        self.layers = {}
        self.reference_layer = None

    def add_layer(self, layer):
        if layer.name is None:
            raise exceptions.UnnamedLayerException(layer, f"Layer name "
                    f"needs to be set for it to be added to {self.name} stack.")
        if layer.name in self.layers:
            raise exceptions.ArgumentError(layer, f"Layer with name "
                    f"'{layer.name}' added twice to '{self.name}' stack.")
        if self.reference_layer is None:
            self.reference_layer = layer
        self.layers[layer.name] = layer

    def remove_layer(self, layer_name):
        del self.layers[layer_name]

    def read_data_dict(self, **index):
        result = {}
        for l in layers:
            result[l.name] = l.read(**index)
        raise NotImplementedError

    def write_data_dict(self, data_dict):
        raise NotImplementedError

    def get_layers(self):
        return list(self.layers.values())

    def get_layers_of_type(self, type_names):
        if isinstance(type_names, str):
            type_names = [type_names]

        types = tuple(str_to_layer_type(n) for n in type_names)

        result = []
        for k, v in six.iteritems(self.layers):
            if isinstance(v, types):
                result.append(v)

        return result

    def get_layer_types(self):
        layer_types = set()
        for k, v in six.iteritems(self.layers):
            layer_types.add(v.get_layer_type())

        return list(layer_types)


class Stack(StackBase):
    def __init__(self, name=None, layer_list=[], folder=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.folder = folder

        for l in layer_list:
            self.add_layer(l)

    def create_sublayer(self, name, layer_type, suffix='', reference=None, **kwargs):
        if self.folder is None:
            raise exceptions.CorgieException("Stack must have 'folder' field set "
                    "before sublayers can be created")

        if self.reference_layer is None and reference is None:
            raise exceptions.CorgieException("Stack must either have at least one layer "
                    "or reference layer must be provided for sublayer creation")

        if reference is None:
            reference = self.reference_layer
        path = os.path.join(self.folder, layer_type, f"{name}{suffix}")
        l = reference.backend.create_layer(path=path, layer_type=layer_type,
                name=name, reference=reference, **kwargs)
        self.add_layer(l)
        return l

    def read_data_dict(self, bcube, mip, translation_adjuster=None, stack_name=None, add_prefix=True):
        data_dict = {}
        field_layers = self.get_layers_of_type("field")
        agg_field = None

        if stack_name is None:
            stack_name == self.name

        if stack_name is None or not add_prefix:
            name_prefix = ""
        else:
            name_prefix = f"{stack_name}_"

        for l in field_layers:
            this_field = l.read(bcube=bcube, mip=mip)
            global_name = "{}{}".format(name_prefix, l.name)
            data_dict[global_name] = this_field

            if agg_field is not None:
                agg_field = compose_fields(agg_field, this_field,
                        is_pix_res=True)
            else:
                agg_field = this_field
        assert (f"{name_prefix}_agg_field" not in data_dict)
        data_dict[f"{name_prefix}agg_field"] = agg_field

        if translation_adjuster is not None:
            translaiton = translation_adjuster(agg_field)
        else:
            translation = helpers.Translation(0, 0)
        final_bcube = copy.deepcopy(bcube)
        final_bcube = final_bcube.translate(
                x_offset=translation.x,
                y_offset=translation.y)

        for l in self.get_layers_of_type(["mask", "img"]):
            global_name = f"{name_prefix}{l.name}"
            data_dict[global_name] = l.read(bcube=bcube, mip=mip)

        return translation, data_dict

    def z_range(self):
        return self.bcube.z_range()

    def cutout(self):
        raise NotImplementedError

def create_stack_from_reference(reference_stack, folder, name, types=None, suffix='', **kwargs):
    result = Stack(name=name, folder=folder)
    if types is None:
        layers = reference_stack.get_layers()
    else:
        layers = reference_stack.get_layers_of_type(types)

    for l in layers:
        result.create_sublayer(name=l.name, layer_type=l.get_layer_type(), suffix=suffix,
                reference=l, dtype=l.get_data_type(), **kwargs)
    return result




import click
import six

from corgie import scheduling, residuals, helpers, stack

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option, \
        create_stack_from_spec

class RenderJob(scheduling.Job):
    def __init__(self, src_stack, dst_stack, mip, pad, render_masks,
                 blackout_masks, seethrough, bcube, chunk_xy, chunk_z,
                 additional_fields=[], seethrough_offset=-1):
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.mip = mip
        self.pad = pad
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.render_masks = render_masks
        self.blackout_masks = blackout_masks
        self.additional_fields = additional_fields
        self.seethrough = seethrough
        self.seethrough_offset = seethrough_offset

        super().__init__()

    def task_generator(self):
        chunks = self.dst_stack.get_layers()[0].break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip=self.mip)

        tasks = [RenderTask(self.src_stack,
                            self.dst_stack,
                            blackout_masks=self.blackout_masks,
                            render_masks=self.render_masks,
                            mip=self.mip,
                            pad=self.pad,
                            bcube=input_chunk,
                            additional_fields=self.additional_fields,
                            seethrough=self.seethrough,
                            seethrough_offset=self.seethrough_offset) \
                                    for input_chunk in chunks]
        corgie_logger.info(f"Yielding render tasks for bcube: {self.bcube}, MIP: {self.mip}")

        yield tasks


class RenderTask(scheduling.Task):
    def __init__(self, src_stack, dst_stack, additional_fields, render_masks,
            blackout_masks, seethrough, seethrough_offset, mip,
            pad, bcube):
        super().__init__(self)
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.render_masks = render_masks
        self.blackout_masks = blackout_masks
        self.mip = mip
        self.bcube = bcube
        self.pad = pad
        self.additional_fields = additional_fields
        self.seethrough = seethrough
        self.seethrough_offset = seethrough_offset
        self.blackout_value = -2.0

    def execute(self):
        padded_bcube = self.bcube.uncrop(self.pad, self.mip)
        seethrough_bcube = self.bcube.translate(
                        z_offset=self.seethrough_offset)

        for f in self.additional_fields:
            self.src_stack.add_layer(f)

        src_translation, src_data_dict = self.src_stack.read_data_dict(padded_bcube,
                mip=self.mip, add_prefix=False)
        agg_field = src_data_dict[f"agg_field"]

        agg_mask = None
        if self.blackout_masks:
            mask_layers = self.dst_stack.get_layers_of_type(["mask"])
            mask_layer_names = [l.name for l in mask_layers]
            for n, d in six.iteritems(src_data_dict):
                if n in mask_layer_names:
                    if agg_mask is None:
                        agg_mask = d
                    else:
                        agg_mask = ((agg_mask + d) > 0).byte()

        if self.render_masks:
            write_layers = self.dst_stack.get_layers_of_type(["img", "mask"])
        else:
            write_layers = self.dst_stack.get_layers_of_type("img")


        for l in write_layers:
            src = src_data_dict[f"{l.name}"]
            if agg_field is not None:
                warped_src = residuals.res_warp_img(src.float(), agg_field)
                if agg_mask is not None:
                    warped_mask = residuals.res_warp_img(
                        agg_mask.float(),
                        agg_field)
                    warped_mask = (warped_mask > 0.4).byte()
                else:
                    warped_mask = None
            else:
                warped_src = src
                warped_mask = agg_mask

            if warped_mask is not None:
                warped_src[warped_mask] = self.blackout_value

            cropped_out = helpers.crop(warped_src, self.pad)

            if l.get_layer_type() == "img" and self.seethrough:
                seethrough_data = l.read(
                        bcube=seethrough_bcube,
                        mip=self.mip)

                seethrough_mask = (cropped_out == 0) + \
                        (cropped_out == self.blackout_value)

                cropped_out[seethrough_mask] = \
                        seethrough_data[seethrough_mask]
                seenthru = (cropped_out[seethrough_mask] != 0).sum()
                corgie_logger.debug(f"Seenthrough {seenthru} pixels")

            l.write(cropped_out, bcube=self.bcube, mip=self.mip)

        for f in self.additional_fields:
            self.src_stack.remove_layer(f.name)


@click.command()
@corgie_optgroup('Layer Parameters')
@corgie_option('--src_layer_spec',  '-s', nargs=1,
        type=str, required=True, multiple=True,
        help='Source layer spec. Use multiple times to include all masks, fields, images. ' + \
                LAYER_HELP_STR)
#
@corgie_option('--dst_folder',  nargs=1,
        type=str, required=True,
        help= "Folder where rendered stack will go")

@corgie_optgroup('Render Method Specification')
@corgie_option('--chunk_xy',       '-c', nargs=1, type=int, default=1024)
@corgie_option('--chunk_z',              nargs=1, type=int, default=1)
@corgie_option('--pad',                  nargs=1, type=int, default=512)
@corgie_option('--mip',                  nargs=1, type=int, required=True)
@corgie_option('--render_masks/--no_render_masks',          default=False)
@corgie_option('--blackout_masks/--no_blackout_masks',      default=True)
@corgie_option('--seethrough/--no_seethrough',          default=True)
@corgie_option('--force_chunk_xy',     is_flag=True)
@corgie_option('--force_chunk_z',      is_flag=True)

@corgie_optgroup('Data Region Specification')
@corgie_option('--start_coord',      nargs=1, type=str, required=True)
@corgie_option('--end_coord',        nargs=1, type=str, required=True)
@corgie_option('--coord_mip',        nargs=1, type=int, default=0)

@click.pass_context
def render(ctx, src_layer_spec, dst_folder, pad, render_masks, blackout_masks,
         seethrough, chunk_xy, chunk_z, start_coord, end_coord, mip,
         coord_mip, force_chunk_xy, force_chunk_z):
    scheduler = ctx.obj['scheduler']

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec,
            name='src', readonly=True)

    if force_chunk_xy:
        force_chunk_xy = chunk_xy
    else:
        force_chunk_xy = None

    if force_chunk_z:
        force_chunk_z = chunk_z
    else:
        force_chunk_z = None

    dst_stack = stack.create_stack_from_reference(reference_stack=src_stack,
            folder=dst_folder, name="dst", types=["img", "mask"],
            force_chunk_xy=force_chunk_xy, force_chunk_z=force_chunk_z, overwrite=True)

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    render_job = RenderJob(src_stack=src_stack,
                           dst_stack=dst_stack,
                           mip=mip,
                           pad=pad,
                           bcube=bcube,
                           chunk_xy=chunk_xy,
                           chunk_z=chunk_z,
                           render_masks=render_masks,
                           blackout_masks=blackout_masks,
                           seethrough=seethrough)

    # create scheduler and execute the job
    scheduler.register_job(render_job, job_name="Render {}".format(bcube))
    scheduler.execute_until_completion()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="corgie",
    version="0.0.9",
    author="Sergiy Popovych",
    author_email="sergiy.popovich@gmail.com",
    description="Connectomics Registration General Inference Engine",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/supersergiy/corgie",
    include_package_data=True,
    package_data={'': ['*.py']},
    install_requires=[
      'sphinx-click',
      'torch',
      'gevent',
      'torchvision',
      'numpy',
      'six',
      'pyyaml',
      'mazepa',
      'click-option-group',
      'click',
      'procspec',
      'modelhouse'
    ],
    entry_points={
        "console_scripts": [
            "corgie = corgie.main:cli",
            "corgie-worker = corgie.worker:worker",
        ],
    },
    packages=setuptools.find_packages(),
    python_requires='>=3.7',
)



import mazepa

wait_until_done = mazepa.Barrier
Task = mazepa.Task
Executor = mazepa.Executor
scheduler_click_options = mazepa.click_options
parse_scheduler_from_kwargs = mazepa.parse_scheduler_from_kwargs
parse_executor_from_kwargs = mazepa.parse_executor_from_kwargs

class Job(mazepa.Job):
    def __init__(self, *kargs, **kwargs):
        self.task_generator = self.task_generator()

    def task_generator():
        raise NotImplemented("Job classes must implement "
                "'task_generator' function")

    def get_tasks(self):
        return next(self.task_generator)

class Scheduler(mazepa.Scheduler):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)

def create_scheduler(*kargs, **kwargs):
    return Scheduler(*kargs, **kwargs)




from gevent import monkey
monkey.patch_all()

import json
import click

from corgie import scheduling
from corgie.log import configure_logger
from corgie.log import logger as corgie_logger

from corgie.argparsers import corgie_option, corgie_optgroup
from corgie.data_backends import get_data_backends, str_to_backend, \
        DEFAULT_DATA_BACKEND, DataBackendBase
from corgie.cli.downsample import downsample

from corgie.cli import get_command_list


class GroupWithCommandOptions(click.Group):
    """ Allow application of options to group with multi command """

    def add_command(self, cmd, name=None):
        click.Group.add_command(self, cmd, name=name)
        # add the group parameters to the command
        for param in self.params:
            cmd.params.append(param)

        # hook the commands invoke with our own
        cmd.invoke = self.build_command_invoke(cmd.invoke)
        self.invoke_without_command = True

    def build_command_invoke(self, original_invoke):

        def command_invoke(ctx):
            """ insert invocation of group function """

            # separate the group parameters
            ctx.obj = dict(_params=dict())
            for param in list(self.params):
                name = param.name
                ctx.obj['_params'][name] = ctx.params[name]
                del ctx.params[name]

            # call the group function with its parameters
            params = ctx.params
            ctx.params = ctx.obj['_params']
            self.invoke(ctx)
            ctx.params = params

            # now call the original invoke (the command)
            original_invoke(ctx)

        return command_invoke


@click.group(cls=GroupWithCommandOptions)
@scheduling.scheduler_click_options
@click.option('--device',     '-d', 'device', nargs=1,
                type=str,
                default='cpu',
                help="Pytorch device specification. Eg: 'cpu', 'cuda', 'cuda:0'",
                show_default=True)
@click.option('-v', '--verbose', count=True, help='Turn on debug logging')
@click.pass_context
def cli(ctx, device, verbose, **kwargs):
    # This little hack let's us make group options look like
    # child command options, and at the same time only execute
    # the setup once
    if ctx.invoked_subcommand is None:
        configure_logger(verbose)
        ctx.obj = {}
        DataBackendBase.default_device = device
        corgie_logger.debug("Creting scheduler...")
        ctx.obj['scheduler'] = scheduling.parse_scheduler_from_kwargs(
                kwargs)

        corgie_logger.debug("Scheduler created.")

for c in get_command_list():
    cli.add_command(c)

if __name__ == "__main__":
    cli()




from copy import deepcopy

COMMAND_LIST = []

from corgie.cli.downsample import downsample
COMMAND_LIST.append(downsample)
from corgie.cli.upsample import upsample
COMMAND_LIST.append(upsample)
from corgie.cli.compute_field import compute_field
COMMAND_LIST.append(compute_field)
from corgie.cli.compute_stats import compute_stats
COMMAND_LIST.append(compute_stats)
from corgie.cli.normalize import normalize
COMMAND_LIST.append(normalize)
from corgie.cli.align_block import align_block
COMMAND_LIST.append(align_block)
from corgie.cli.render import render
COMMAND_LIST.append(render)
from corgie.cli.copy import copy
COMMAND_LIST.append(copy)
from corgie.cli.apply_processor import apply_processor
COMMAND_LIST.append(apply_processor)

# To add new commands, create a file in this folder implementing a command,
# import the command here and add it to the list:


def get_command_list():
    return COMMAND_LIST



from gevent import monkey
monkey.patch_all()

import click

from corgie import scheduling
from corgie.log import logger as corgie_logger
from corgie.log import configure_logger


@click.command()
@click.option('--lease_seconds', '-l', nargs=1, type=int, required=True)
@scheduling.scheduler_click_options
@click.option('-v', '--verbose', count=True, help='Turn on debug logging')
def worker(lease_seconds, verbose, **kwargs):
    configure_logger(verbose)
    executor = scheduling.parse_executor_from_kwargs(kwargs)
    executor.execute(lease_seconds=lease_seconds)



import click
import procspec

from corgie import scheduling, argparsers, helpers

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option, \
        create_stack_from_spec

from corgie.cli.common import ChunkedJob


class ComputeFieldJob(scheduling.Job):
    def __init__(self, src_stack, tgt_stack, dst_layer,
            chunk_xy, chunk_z, processor_spec, processor_mip,
            pad, crop,
            bcube, tgt_z_offset, blend_xy=0, suffix=''):

        self.src_stack = src_stack
        self.tgt_stack = tgt_stack
        self.dst_layer = dst_layer
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.blend_xy = blend_xy
        self.pad = pad
        self.crop = crop
        self.bcube = bcube
        self.tgt_z_offset = tgt_z_offset
        self.suffix = suffix #in case this job wants to make more layers
        self.processor_spec = processor_spec
        self.processor_mip = processor_mip
        if isinstance(self.processor_spec, str):
            self.processor_spec = [self.processor_spec]
        if isinstance(self.processor_mip, int):
            self.processor_mip = [self.processor_mip]

        if len(self.processor_mip) != len(self.processor_spec):
            raise exceptions.CorgieException("The number of processors doesn't "
                    "match the number of specified processor MIPs")

        super().__init__()

    def task_generator(self):
        for i in range(len(self.processor_spec)):
            this_proc = self.processor_spec[i]
            this_proc_mip = self.processor_mip[i]
            is_last_proc = i == len(self.processor_spec) - 1

            this_task = helpers.PartialSpecification(
                    ComputeFieldTask,
                    src_stack=self.src_stack,
                    tgt_stack=self.tgt_stack,
                    processor_spec=this_proc,
                    pad=self.pad,
                    crop=self.crop,
                    tgt_z_offset=self.tgt_z_offset
                    )

            chunked_job = ChunkedJob(
                    task_class=this_task,
                    dst_layer=self.dst_layer,
                    chunk_xy=self.chunk_xy,
                    chunk_z=self.chunk_z,
                    blend_xy=self.blend_xy,
                    mip=this_proc_mip,
                    bcube=self.bcube,
                    suffix=self.suffix
                    )


            yield from chunked_job.task_generator

            if not is_last_proc:
                yield scheduling.wait_until_done
                next_proc_mip = self.processor_mip[i + 1]
                if this_proc_mip > next_proc_mip:
                    downsample_job = DownsampleJob(
                                src_layer=self.dst_layer,
                                chunk_xy=self.chunk_xy,
                                chunk_z=self.chunk_z,
                                mip_start=this_proc_mip,
                                mip_end=next_proc_mip,
                                bcube=self.bcube
                                )
                    yield from downsample_job.task_generator
                    yield scheduling.wait_until_done


        if self.processor_mip[0] > self.processor_mip[-1]:
            # good manners
            # prepare the ground for the next you
            downsample_job = DownsampleJob(
                            src_layer=self.dst_layer,
                            chunk_xy=self.chunk_xy,
                            chunk_z=self.chunk_z,
                            mip_start=self.processor_mip[-1],
                            mip_end=self.processor_mip[0],
                            bcube=self.bcube
                            )


class ComputeFieldTask(scheduling.Task):
    def __init__(self, processor_spec, src_stack, tgt_stack, dst_layer,  mip,
            pad, crop, tgt_z_offset, bcube):
        super().__init__()
        self.processor_spec = processor_spec
        self.src_stack = src_stack
        self.tgt_stack = tgt_stack
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.crop = crop
        self.tgt_z_offset = tgt_z_offset
        self.bcube = bcube

    def execute(self):
        src_bcube = self.bcube.uncrop(self.pad, self.mip)
        tgt_bcube = src_bcube.translate(z_offset=self.tgt_z_offset)

        processor = procspec.parse_proc(
                spec_str=self.processor_spec)

        src_translation, src_data_dict = self.src_stack.read_data_dict(src_bcube,
                mip=self.mip, stack_name='src')
        tgt_translation, tgt_data_dict = self.tgt_stack.read_data_dict(tgt_bcube,
                mip=self.mip, stack_name='tgt')

        processor_input = {**src_data_dict, **tgt_data_dict}

        predicted_field = processor(processor_input,
                                    output_key='src_cf_field')

        cropped_field = helpers.crop(predicted_field, self.crop)
        cropped_field[:, 0] += src_translation.x
        cropped_field[:, 1] += src_translation.y
        self.dst_layer.write(cropped_field, bcube=self.bcube, mip=self.mip)


@click.command()

@corgie_optgroup('Layer Parameters')

@corgie_option('--src_layer_spec',  '-s', nargs=1,
        type=str, required=True, multiple=True,
        help='Source layer spec. Use multiple times to include all masks, fields, images. ' + \
                LAYER_HELP_STR)
#
@corgie_option('--tgt_layer_spec', '-t', nargs=1,
        type=str, required=False, multiple=True,
        help='Target layer spec. Use multiple times to include all masks, fields, images. '\
                'DEFAULT: Same as source layers')

@corgie_option('--dst_layer_spec',  nargs=1,
        type=str, required=False,
        help= "Specification for the destination layer. Must be a field type." + \
                " DEFAULT: source reference key path + /field/cf_field + (_{suffix})?")

@corgie_option('--reference_key',        nargs=1, type=str, default='img')

@corgie_optgroup('Compute Field Method Specification')
@corgie_option('--chunk_xy',       '-c', nargs=1, type=int, default=1024)
@corgie_option('--chunk_z',              nargs=1, type=int, default=1)
@corgie_option('--blend_xy',             nargs=1, type=int, default=0)
@corgie_option('--pad',                  nargs=1, type=int, default=512,
        )
@corgie_optgroup('')
@corgie_option('--crop',                 nargs=1, type=int, default=None)
@corgie_option('--processor_spec',       nargs=1, type=str, multiple=True,
        required=True)
@corgie_option('--processor_mip',                  nargs=1, type=int, multiple=True,
        required=True)

@corgie_optgroup('Data Region Specification')
@corgie_option('--start_coord',      nargs=1, type=str, required=True)
@corgie_option('--end_coord',        nargs=1, type=str, required=True)
@corgie_option('--coord_mip',        nargs=1, type=int, default=0)
@corgie_option('--tgt_z_offset',     nargs=1, type=str, default=1)

@click.option('--suffix',            nargs=1, type=str,  default=None)

@click.pass_context
def compute_field(ctx, src_layer_spec, tgt_layer_spec, dst_layer_spec,
        suffix, processor_spec, pad, crop, chunk_xy, start_coord, processor_mip,
        end_coord, coord_mip, blend_xy, tgt_z_offset, chunk_z, reference_key):
    if suffix is None:
        suffix = ''
    else:
        suffix = f"_{suffix}"

    scheduler = ctx.obj['scheduler']

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec,
            name='src', readonly=True)

    tgt_stack = create_stack_from_spec(tgt_layer_spec,
            name='tgt', readonly=True, reference=src_stack)

    reference_layer = None
    if reference_key in src_stack.layers:
        reference_layer = src_stack.layers[reference_key]
    dst_layer = create_layer_from_spec(dst_layer_spec, allowed_types=['field'],
            default_type='field', readonly=False, caller_name='dst_layer',
            reference=reference_layer, overwrite=True)

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if crop is None:
        crop = pad

    compute_field_job = ComputeFieldJob(
            src_stack=src_stack,
            tgt_stack=tgt_stack,
            dst_layer=dst_layer,
            chunk_xy=chunk_xy,
            chunk_z=chunk_z,
            blend_xy=blend_xy,
            processor_spec=processor_spec,
            pad=pad,
            crop=crop,
            bcube=bcube,
            tgt_z_offset=tgt_z_offset,
            suffix=suffix,
            processor_mip=processor_mip)

    # create scheduler and execute the job
    scheduler.register_job(compute_field_job, job_name="Compute field {}, tgt z offset {}".format(
        bcube, tgt_z_offset))
    scheduler.execute_until_completion()



import click
import torch
import procspec
import cachetools

from corgie import scheduling, argparsers, helpers

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, \
                             str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import LAYER_HELP_STR, \
        create_layer_from_spec, corgie_optgroup, corgie_option, \
        create_stack_from_spec

@cachetools.cached(cachetools.LRUCache(maxsize=4))
def get_gaussian_mask(shape, var):
    z = shape[0]
    channels = shape[1]
    kernel_size = shape[-2:]

    kernel = 1
    s = 150.0 * kernel_size[0] / 1024
    sigma = [s, s]
    meshgrids = torch.meshgrid(
            [
                torch.arange(size, dtype=torch.float32)
                        for size in kernel_size
            ]
        )

    for size, std, mgrid in zip(kernel_size, sigma, meshgrids):
        mean = (size - 1) / 2
        kernel *= torch.exp(-((mgrid - mean) / std) ** 2 / 2)

    kernel = kernel.view(1, 1, *kernel.size())
    kernel = kernel.repeat(z, channels, 1, 1)
    return kernel


class BlendDivideTask(scheduling.Task):
    def __init__(self, dst_layer, accum_layer, weight_layer, count_layer,
             mip, bcube):
        self.dst_layer = dst_layer
        self.accum_layer = accum_layer
        self.weight_layer = weight_layer
        self.count_layer = count_layer
        self.mip = mip
        self.bcube = bcube
        super().__init__()

    def execute(self):
        accum_data = self.accum_layer.read(bcube=self.bcube, mip=self.mip)
        weight_data = self.weight_layer.read(bcube=self.bcube, mip=self.mip)
        count_data = self.count_layer.read(bcube=self.bcube, mip=self.mip)

        dst_data = accum_data
        many_writes_mask = (count_data > 1).squeeze()
        dst_data[..., many_writes_mask] /= \
                weight_data[..., many_writes_mask]

        self.dst_layer.write(dst_data, bcube=self.bcube, mip=self.mip)


class BlendAccumulateTask(scheduling.Task):
    def __init__(self, src_layer, accum_layer, weight_layer, count_layer,
            blend_var, mip, bcube):
        self.src_layer = src_layer
        self.accum_layer = accum_layer
        self.weight_layer = weight_layer
        self.count_layer = count_layer
        self.blend_var = blend_var
        self.mip = mip
        self.bcube = bcube
        super().__init__()

    def execute(self):
        src_data = self.src_layer.read(bcube=self.bcube, mip=self.mip)
        accum_data = self.accum_layer.read(bcube=self.bcube, mip=self.mip)
        weight_data = self.weight_layer.read(bcube=self.bcube, mip=self.mip)
        count_data = self.count_layer.read(bcube=self.bcube, mip=self.mip)

        weight_mask = get_gaussian_mask(src_data.shape, self.blend_var)

        first_write_mask = (count_data == 0).squeeze()
        second_write_mask = (count_data == 1).squeeze()

        accum_data[..., first_write_mask] = \
                src_data[..., first_write_mask]
        accum_data[..., second_write_mask] *= \
                weight_data[..., second_write_mask]

        accum_data[..., first_write_mask == False] += \
                weight_mask[..., first_write_mask == False] * \
                src_data[..., first_write_mask == False]

        # dst_layer may have many channels,
        # but our weights will have only one since it's the
        # same for channels
        weight_data += weight_mask[:, 0:1]

        count_data += 1

        self.accum_layer.write(accum_data, bcube=self.bcube, mip=self.mip)
        self.weight_layer.write(weight_data, bcube=self.bcube, mip=self.mip)
        self.count_layer.write(count_data, bcube=self.bcube, mip=self.mip)


class ChunkedJob(scheduling.Job):
    ''' Job that applies the given task to the bcube in chunk.
        Supports chunk blening'''

    def __init__(self, task_class, dst_layer, chunk_xy, chunk_z, bcube, mip,
            blend_xy=0, blend_var=1,
            suffix=''):
        ''' blend_xy decides the blend region.
            it must be less than 50% of chunk_xy

            task_class is PartialSpecification of a task
            lacking only destination layer, bcube and MIP
            parameters'''
        self.suffix = suffix
        self.task_class = task_class

        self.dst_layer = dst_layer
        self.bcube = bcube
        self.mip = mip

        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z

        if blend_xy < 0:
            raise exceptions.CorgieException(f"'blend_xy' must "
                    f"be a positive integer. Received: {blend_xy}")

        if blend_xy > chunk_xy // 2:
            raise exceptions.CorgieException(f"'blend_xy' must "
                    f"be less than half of 'chunk_xy'. Given:"
                    f"'blend_xy': {blend_xy}, 'chunk_xy': {chunk_xy}")

        self.blend_xy = blend_xy
        self.blend_var = blend_var
        super().__init__()

    def task_generator(self):
        if self.blend_xy == 0:
            yield from self.noblend_task_generator()
        else:
            yield from self.blend_task_generator()

    def noblend_task_generator(self):
        chunks = self.dst_layer.break_bcube_into_chunks(
                    bcube=self.bcube,
                    chunk_xy=self.chunk_xy,
                    chunk_z=self.chunk_z,
                    mip=self.mip)

        tasks = [self.task_class(dst_layer=self.dst_layer,
                                 mip=self.mip,
                                 bcube=chunk) for chunk in chunks]

        corgie_logger.debug(f"Yielding {type(tasks[0])} tasks "\
                f"for bcube: {self.bcube}, MIP: {self.mip}")
        yield tasks

    def blend_task_generator(self):
        assert(self.chunk_xy % self.blend_xy == 0)

        accum_layer  = self.dst_layer.get_sublayer(
                name=f"accum_blend{self.suffix}",
                layer_type=self.dst_layer.get_layer_type(),
                force_chunk_xy=self.blend_xy,
                force_chunk_z=self.chunk_z,
                overwrite=True)
        weight_layer = self.dst_layer.get_sublayer(
                name=f"accum_weight{self.suffix}",
                force_chunk_xy=self.blend_xy,
                force_chunk_z=self.chunk_z,
                layer_type='img', dtype='float32',
                overwrite=True)
        count_layer  = self.dst_layer.get_sublayer(
                name=f"accum_count{self.suffix}",
                force_chunk_xy=self.blend_xy,
                force_chunk_z=self.chunk_z,
                layer_type='img', dtype='uint8',
                overwrite=True)

        blend_chunks = accum_layer.break_bcube_into_chunks(
                    bcube=self.bcube,
                    chunk_xy=self.chunk_xy,
                    chunk_z=self.chunk_z,
                    chunk_xy_step=self.chunk_xy - self.blend_xy,
                    mip=self.mip,
                    flatten=False)

        checkerb_layers, checkerb_chunk_sets = self._create_checkerboard(blend_chunks)
        for i in range(len(checkerb_chunk_sets)):
            tasks = [self.task_class(dst_layer=checkerb_layers[i],
                                     mip=self.mip,
                                     bcube=chunk) \
                                 for chunk in checkerb_chunk_sets[i]]

            corgie_logger.debug(f"Yielding {type(tasks[0])} tasks "\
                    f"for bcube: {self.bcube}, MIP: {self.mip}, "
                    f"checkerboard #{i}")
            yield tasks

        yield scheduling.wait_until_done

        for i in range(len(checkerb_chunk_sets)):
            tasks = [BlendAccumulateTask(src_layer=checkerb_layers[i],
                                         accum_layer=accum_layer,
                                         weight_layer=weight_layer,
                                         count_layer=count_layer,
                                         blend_var=self.blend_var,
                                         mip=self.mip,
                                         bcube=chunk) \
                                 for chunk in checkerb_chunk_sets[i]]

            corgie_logger.debug(f"Yielding {type(tasks[0])} tasks "\
                    f"for bcube: {self.bcube}, MIP: {self.mip}, "
                    f"checkerboard #{i}")

            yield tasks
            yield scheduling.wait_until_done

        dst_chunks = self.dst_layer.break_bcube_into_chunks(
                    bcube=self.bcube,
                    chunk_xy=self.chunk_xy,
                    chunk_z=self.chunk_z,
                    mip=self.mip)

        tasks = [BlendDivideTask(dst_layer=self.dst_layer,
                                 accum_layer=accum_layer,
                                 weight_layer=weight_layer,
                                 count_layer=count_layer,
                                 mip=self.mip,
                                 bcube=chunk) \
                                     for chunk in dst_chunks]
        yield tasks


    def _create_checkerboard(self, blend_chunks):
        checkerb_layers = []
        checkerb_chunk_sets = []
        count = 0

        for x_chunk_offset in [0, 1]:
            for y_chunk_offset in [0, 1]:
                checkerb_layer = self.dst_layer.get_sublayer(
                        name=f"checkerboard_{count}{''}",
                        layer_type=self.dst_layer.get_layer_type(),
                        force_chunk_xy=self.blend_xy,
                        force_chunk_z=self.chunk_z,
                        overwrite=True
                        )

                checkerb_chunks = self._get_checkerboard_chunks(
                                                    xy_chunks=blend_chunks,
                                                    x_offset=x_chunk_offset,
                                                    y_offset=y_chunk_offset)

                checkerb_layers.append(checkerb_layer)
                checkerb_chunk_sets.append(checkerb_chunks)
                count += 1

        return checkerb_layers, checkerb_chunk_sets

    def _get_checkerboard_chunks(self, xy_chunks, x_offset, y_offset):
        result = []
        for z in range(len(xy_chunks)):
            for x in range(x_offset, len(xy_chunks[0]), 2):
                for y in range(y_offset, len(xy_chunks[0][0]), 2):
                    result.append(xy_chunks[z][x][y])
        return result




